# -*- coding: utf-8 -*-

##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################

"""
I/O utilities and domain helper functions for escript.

This module provides:
- Constants: EPSILON, DBLE_MAX, getEpsilon, getMaxFloat
- Data I/O: interpolateTable, saveDataCSV, getNumpy, convertToNumpy, NumpyToData, saveESD
- Domain utilities: getTagNames, insertTagNames, insertTaggedValues, diameter,
  boundingBoxEdgeLengths, boundingBox, getBoundingBox, longestEdge, getRegionTags, makeTagMap
- Filesystem: mkDir, FileWriter class, BBxInterval class

:var __author__: name of author
:var __copyright__: copyrights
:var __license__: licence agreement
:var __url__: url entry point on documentation
:var EPSILON: smallest positive value with 1.<1.+EPSILON
:var DBLE_MAX: largest positive float
"""


__copyright__="""Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__="""Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""
__url__="https://github.com/LutzGross/esys-escript.github.io"
__author__="Lutz Gross, l.gross@uq.edu.au"

import math
import os
import warnings
import numpy

from . import escriptcpp as escore
from .escriptcpp import Data, getMPIRankWorld, getMPIWorldMax, hasFeature


#=========================================================
#   Constants and helpers
#=========================================================
def getEpsilon():
     """
     Returns the machine epsilon (smallest positive value where 1.0 + epsilon != 1.0).

     :return: machine precision epsilon
     :rtype: ``float``
     """
     return escore.getMachinePrecision()
EPSILON=getEpsilon()

def getMaxFloat():
     """
     Returns the largest representable positive floating point number.

     :return: maximum float value
     :rtype: ``float``
     """
     return escore.getMaxFloat()
DBLE_MAX=getMaxFloat()


#=========================================================
#   Tag utilities
#=========================================================
def getTagNames(domain):
    """
    Returns a list of tag names used by the domain.

    :param domain: a domain object
    :type domain: `escript.Domain`
    :return: a list of tag names used by the domain
    :rtype: ``list`` of ``str``
    """
    return [n.strip() for n in domain.showTagNames().split(",") ]

def insertTagNames(domain,**kwargs):
    """
    Inserts tag names into the domain.

    :param domain: a domain object
    :type domain: ``escript.Domain``
    :keyword <tag_name>: tag key assigned to <tag_name>
    :type <tag_name>: ``int``
    """
    for  k in kwargs:
         domain.setTagMap(k,kwargs[k])

def insertTaggedValues(target,**kwargs):
    """
    Inserts tagged values into the target using tag names.

    :param target: data to be filled by tagged values
    :type target: `escript.Data`
    :keyword <tag_name>: value to be used for <tag_name>
    :type <tag_name>: ``float`` or ``numpy.ndarray``
    :return: ``target``
    :rtype: `escript.Data`
    """
    for k in kwargs:
        target.setTaggedValue(k,kwargs[k])
    return target

def getRegionTags(function_space):
    """
    Returns the tag distribution of a function_space as a Data object.

    :param function_space: function_space to be used
    :type function_space: `escript.FunctionSpace`
    :return: a data object which contains the tag distribution
    :rtype: `escript.Data` of rank 0 with ReducedFunction attributes.
    """

    out = escore.Scalar(0., function_space)
    for tag in function_space.getListOfTags():
        out.setTaggedValue(tag,float(tag))
    out.expand()
    return out

def makeTagMap(fs):
    """
    Produces an expanded Data over the function space where
    the value is the tag associated with the sample.

    :param fs: the function space
    :type fs: `escript.FunctionSpace`
    :return: Data object with tag values at each sample point
    :rtype: `escript.Data`
    """
    out=escore.Scalar(0, fs)    # The default tag is zero anyway
    for t in fs.getListOfTags():
        out.setTaggedValue(t,t)
    out.expand()
    return out


#=========================================================
#   Data I/O functions
#=========================================================
def interpolateTable(tab, dat, start, step, undef=1.e50, check_boundaries=False):
    """
    Interpolates values from a lookup table.

    .. deprecated::
       Use :class:`~esys.escriptcore.interpolation.InterpolationTable` instead.
       This function delegates to ``InterpolationTable`` with ``order=1``.

    :param tab: the lookup table array
    :type tab: ``numpy.ndarray``
    :param dat: coordinate data — scalar `Data` for 1-D or vector `Data` for 2-D/3-D
    :type dat: `Data`
    :param start: starting coordinate(s) for the table
    :type start: ``float`` or ``tuple`` of ``float``
    :param step: step size(s) for the table
    :type step: ``float`` or ``tuple`` of ``float``
    :param undef: upper bound on interpolated values
    :type undef: ``float``
    :param check_boundaries: if ``True``, raise an error for out-of-bounds coordinates
    :type check_boundaries: ``bool``
    :return: interpolated values
    :rtype: `Data`
    """
    warnings.warn(
        "interpolateTable is deprecated; use InterpolationTable instead.",
        DeprecationWarning, stacklevel=2)
    from .interpolation import InterpolationTable
    return InterpolationTable(tab, start, step, order=1, undef=undef,
                              check_boundaries=check_boundaries,
                              table_indexing="kji")(dat)


def saveDataCSV(filename, append=False, refid=False, sep=", ", csep="_", **data):
    """
    Writes `Data` objects to a CSV file.
    These objects must have compatible FunctionSpaces, i.e. it must be possible
    to interpolate all data to one `FunctionSpace`. Note, that with more than
    one MPI rank this function  will fail for some function spaces on some
    domains.

    :param filename: file to save data to.
    :type filename: ``string``
    :param append: If ``True``, then open file at end rather than beginning
    :type append: ``bool``
    :param refid: If ``True``, then a list of reference ids will be printed in the first column
    :type refid: ``bool``
    :param sep: separator between fields
    :type sep: ``string``
    :param csep: separator for components of rank 2 and above (e.g. '_' -> c0_1)

    The keyword args are Data objects to save.
    If a scalar `Data` object is passed with the name ``mask``, then only
    samples which correspond to positive values in ``mask`` will be output.
    Example::

        s=Scalar(..)
        v=Vector(..)
        t=Tensor(..)
        f=float()
        saveDataCSV("f.csv", a=s, b=v, c=t, d=f)

    Will result in a file

    a, b0, b1, c0_0, c0_1, .., c1_1, d
    1.0, 1.5, 2.7, 3.1, 3.4, .., 0.89,  0.0
    0.9, 8.7, 1.9, 3.4, 7.8, .., 1.21,  0.0

    The first line is a header, the remaining lines give the values.

    """
    # find a function space:
    fs = None
    for n,d in sorted(data.items(), key=lambda x: x[0]):
        if isinstance(d, Data): fs=d.getFunctionSpace()
    if fs is None:
        raise ValueError("saveDataCSV: there must be at least one Data object in the argument list.")

    new_data={}
    for n,d in sorted(data.items(), key=lambda x: x[0]):
        if isinstance(d, Data):
            new_data[n]=d
        else:
            try:
               new_data[n]=Data(d,fs)
            except:
               raise ValueError("saveDataCSV: unknown non-data argument type for %s"%(str(n)))
    escore._saveDataCSV(filename, new_data, sep, csep, refid, append)


def getNumpy(**data):
    """
    Converts `Data` objects to numpy arrays.

    The keyword args are Data objects to convert.
    If a scalar `Data` object is passed with the name ``mask``, then only
    samples which correspond to positive values in ``mask`` will be output.

    :keyword <name>: Data object or scalar to convert
    :type <name>: `Data` or ``float``
    :return: dictionary mapping names to numpy record arrays
    :rtype: ``dict`` of ``numpy.ndarray``
    :raise ValueError: if no Data object is provided

    Example usage::

        s=Scalar(..)
        v=Vector(..)
        t=Tensor(..)
        f=float()
        array = getNumpy(a=s, b=v, c=t, d=f)
    """
    # find a function space:
    fs = None
    for n,d in sorted(data.items(), key=lambda x: x[0]):
        if isinstance(d, Data): fs=d.getFunctionSpace()
    if fs is None:
        raise ValueError("getNumpy: there must be at least one Data object in the argument list.")

    new_data={}
    for n,d in sorted(data.items(), key=lambda x: x[0]):
        if isinstance(d, Data):
            new_data[n]=d
        else:
            try:
                new_data[n]=Data(d,fs)
            except:
                raise ValueError("getNumpy: unknown non-data argument type for %s"%(str(n)))

    numpy_arrays = escore._getNumpy(new_data)

    # Get some information about the shape of the data
    counter = -1
    shape=[]
    names=[]
    index=[]
    for i in range(0, len(numpy_arrays)):
        tmp = numpy_arrays[i]
        # New variable?
        if isinstance(tmp, str) is True:
            name = tmp
            counter+=1
            shape.append(0)
            names.append(name)
            index.append(i+1)
        else:
            shape[counter]+=1

    # Dictionary to store the information in
    answer = {}
    for i in range(0, len(names)):
        if(shape[i] == 1):
            answer.update({names[i] : numpy.rec.fromarrays(numpy_arrays[index[i]],names=names[i])})
        else:
            tmp=numpy.asarray(numpy.rec.fromarrays(numpy_arrays[index[i]],names=names[i]))
            for j in range(1, shape[i]):
                tmp=numpy.hstack([tmp,numpy.rec.fromarrays(numpy_arrays[index[i]]+j,names=names[i])])
            answer.update({names[i] : tmp})

    return answer

def convertToNumpy(data):
    """
    Converts a `Data` object to a numpy array.

    :param data: the Data object to convert
    :type data: `Data`
    :return: numpy array containing the data values
    :rtype: ``numpy.ndarray``
    """
    if hasFeature("boostnumpy"):
        return escore._convertToNumpy(Data(data,data.getFunctionSpace()))
    else:
        N=data.getNumberOfDataPoints()
        if data.getShape() == ():
            S = ( 1 , N )
        else:
            S = data.getShape() + (N,)
        if data.isComplex():
            out = numpy.empty(S, dtype = numpy.complex128)
        else:
            out = numpy.empty( S, dtype = numpy.float64)

        for k in range(data.getNumberOfDataPoints()):
            out[:,k] = numpy.array( data.getTupleForDataPoint(k), dtype = out.dtype, subok = False, ndmin = 1 )
        return out


def NumpyToData(array, isComplex, functionspace):
    """
    Creates a `Data` object from a numpy ndarray.

    :param array: the numpy array to convert
    :type array: ``numpy.ndarray``
    :param isComplex: whether the data contains complex values
    :type isComplex: ``bool``
    :param functionspace: the function space for the new Data object
    :type functionspace: `FunctionSpace`
    :return: Data object created from the numpy array
    :rtype: `Data`
    :raise ValueError: if arguments are invalid or boost numpy is not available

    Example usage::

        NewDataObject = NumpyToData(ndarray, isComplex, FunctionSpace)
    """
    if hasFeature("boostnumpy"):
      if(not isinstance(array, (numpy.ndarray, numpy.generic))):
        raise ValueError("NumpyToData: Invalid argument for array.")
      if(not isinstance(functionspace, escore.FunctionSpace)):
        raise ValueError("NumpyToData: Invalid argument for functionspace.")
      # escore.internal_initBoostNumpy();
      return escore._numpyToData(array, isComplex, functionspace)
    else:
      raise ValueError("NumpyToData: Please recompile escript with boost numpy.")


def saveESD(datasetName, dataDir=".", domain=None, timeStep=0, deltaT=1, dynamicMesh=0, timeStepFormat="%04d", **data):
    """
    Saves `Data` objects to files and creates an `escript dataset` (ESD) file
    for convenient processing/visualisation.

    Single timestep example::

        tmp = Scalar(..)
        v = Vector(..)
        saveESD("solution", "data", temperature=tmp, velocity=v)

    Time series example::

        while t < t_end:
            tmp = Scalar(..)
            v = Vector(..)
            # save every 10 timesteps
            if t % 10 == 0:
                saveESD("solution", "data", timeStep=t, deltaT=10, temperature=tmp, velocity=v)
            t = t + 1

    tmp, v and the domain are saved in native format in the "data"
    directory and the file "solution.esd" is created that refers to tmp by
    the name "temperature" and to v by the name "velocity".

    :param datasetName: name of the dataset, used to name the ESD file
    :type datasetName: ``str``
    :param dataDir: optional directory where the data files should be saved
    :type dataDir: ``str``
    :param domain: domain of the `Data` object(s). If not specified, the
                   domain of the given `Data` objects is used.
    :type domain: `escript.Domain`
    :param timeStep: current timestep or sequence number - first one must be 0
    :type timeStep: `int`
    :param deltaT: timestep or sequence increment, see example above
    :type deltaT: `int`
    :param dynamicMesh: by default the mesh is assumed to be static and thus
                        only saved once at timestep 0 to save disk space.
                        Setting this to 1 changes the behaviour and the mesh
                        is saved at each timestep.
    :type dynamicMesh: `int`
    :param timeStepFormat: timestep format string (defaults to "%04d")
    :type timeStepFormat: ``str``
    :keyword <name>: writes the assigned value to the file using <name> as
                     identifier
    :type <name>: `Data` object
    :note: The ESD concept is experimental and the file format likely to
           change so use this function with caution.
    :note: The data objects have to be defined on the same domain (but not
           necessarily on the same `FunctionSpace`).
    :note: When saving a time series the first timestep must be 0 and it is
           assumed that data from all timesteps share the domain. The dataset
           file is updated in each iteration.
    """
    # Local import to avoid circular dependency
    from . import util

    new_data = {}
    for n,d in sorted(data.items(), key=lambda x: x[0]):
          if not d.isEmpty():
            fs = d.getFunctionSpace()
            domain2 = fs.getDomain()
            if fs == escore.Solution(domain2):
               new_data[n]=util.interpolate(d,escore.ContinuousFunction(domain2))
            elif fs == escore.ReducedSolution(domain2):
               new_data[n]=util.interpolate(d,escore.ReducedContinuousFunction(domain2))
            else:
               new_data[n]=d
            if domain is None: domain=domain2
    if domain is None:
        raise ValueError("saveESD: no domain detected.")

    if domain.isRootRank() and not os.path.isdir(dataDir):
        os.mkdir(dataDir)

    meshFile = datasetName+"_mesh"
    fileNumber = timeStep / deltaT

    if dynamicMesh == 0:
        # later timesteps reuse mesh from t=0
        if timeStep == 0:
            domain.dump(os.path.join(dataDir, meshFile))
    else:
        meshFile += ".%s"%timeStepFormat
        domain.dump(os.path.join(dataDir, (meshFile) % fileNumber))

    outputString = ""

    if domain.isRootRank():
        outputString += "#escript datafile V1.0\n"
        # number of timesteps (currently only 1 is supported)
        outputString += "T=%d\n" % (fileNumber+1)
        # timestep increment
        outputString += "DT=1\n"
        # name of the mesh file
        outputString += "M=%s\n" % meshFile
        # number of blocks (MPI size)
        outputString += "N=%d\n" % domain.getMPISize()

    # now add the variables
    for varName,d in sorted(new_data.items(), key=lambda x: x[0]):
        varFile = datasetName+"_"+varName+".%s"%timeStepFormat
        d.dump(os.path.join(dataDir, (varFile ) % fileNumber))
        if domain.isRootRank():
            outputString += "V=%s:%s\n" % (varFile, varName)

    if domain.isRootRank():
        esdfile = open(os.path.join(dataDir, datasetName+".esd"), "w")
        esdfile.write(outputString)
        esdfile.close()


#=========================================================
#   Domain utilities
#=========================================================
def diameter(domain):
    """
    Returns the diameter of a domain.

    :param domain: a domain
    :type domain: `escript.Domain`
    :rtype: ``float``
    """
    return math.sqrt(sum( [ v**2 for v in boundingBoxEdgeLengths(domain) ] ))

def boundingBoxEdgeLengths(domain):
    """
    Returns the edge lengths of the bounding box of a domain

    :param domain: a domain
    :type domain: `escript.Domain`
    :rtype: ``list`` of ``float``
    """
    return  [ v[1]-v[0] for v in boundingBox(domain) ]


def boundingBox(domain):
    """
    Returns the bounding box of a domain

    :param domain: a domain
    :type domain: `escript.Domain`
    :return: bounding box of the domain
    :rtype: ``list`` of pairs of ``float``
    """
    # Local import to avoid circular dependency
    from . import util

    x=domain.getX()
    out=[]
    for i in range(domain.getDim()):
       x_i=x[i]
       out.append((util.inf(x_i), util.sup(x_i)))
    return out

class BBxInterval(object):
    """
    Represents an interval for bounding box coordinates.
    """
    def __init__(self, minval=0, maxval=0):
        self.min=minval
        self.max=maxval
    def __str__(self):
        return f"[{self.min}, {self.max}]"
    def __repr__(self):
        return f"[{self.min}, {self.max}]"

def getBoundingBox(domain):
    """
    Returns the bounding box of a domain as a list of intervals

    :param domain: a domain
    :type domain: `escript.Domain`
    :return: bounding box of the domain
    :rtype: ``list`` of ``BBxInterval``
    """
    return [ BBxInterval(minval, maxval) for minval, maxval in boundingBox(domain)]

def longestEdge(domain):
    """
    Returns the length of the longest edge of the domain

    :param domain: a domain
    :type domain: `escript.Domain`
    :return: longest edge of the domain parallel to the Cartesian axis
    :rtype: ``float``
    """
    return max(boundingBoxEdgeLengths(domain))


#=========================================================
#   Filesystem utilities
#=========================================================
def mkDir(*pathname):
    """
    creates a directory of name ``pathname`` if the directory does not exist.

    :param pathname: valid path name
    :type pathname: ``str`` or ``sequence of strings``
    :note: The method is MPI safe.
    """
    errno = 0
    p_fail = None
    ex = None
    if getMPIRankWorld() == 0:
      for p in pathname:
       if os.path.exists(p):
          if not os.path.isdir(p):
                errno = 2
                p_fail = p
       else:
          try:
              os.makedirs(p)
          except Exception as e:
              ex = e
              errno = 1
              p_fail = p

    errno = getMPIWorldMax(errno)
    if errno > 0:
         if errno==2:
            if p_fail is None:
               raise IOError("Unable to create directory.")
            else:
               raise IOError("Unable to create directory %s. It already exists and is not a directory."%p_fail)
         elif ex is None:
            if p_fail is None:
               raise IOError("Unable to create directory.")
            else:
               raise IOError("Unable to create directory %s."%p_fail)
         else:
            if len(str(ex)) > 0:
               raise IOError(str(ex))
            else:
               if p_fail is None:
                  raise IOError("Unable to create directory.")
               else:
                  raise IOError("Unable to create directory %s."%p_fail)


class FileWriter(object):
    """
    Interface to write data to a file. In essence this class wraps the standard ``file`` object to write data that are global in MPI
    to a file. In fact, data are written on the processor with MPI rank 0 only. It is recommended to use ``FileWriter`` rather than ``open`` in order to write
    code that is running with as well as without MPI. It is safe to use ``open`` under MPI to read data which are global under MPI.

    :var name: name of file
    :var mode: access mode (='w' or ='a')
    :var closed: True to indicate closed file
    :var newlines: line separator
    """
    def __init__(self,fn,append=False,createLocalFiles=False):
         """
         Opens a file of name ``fn`` for writing. If running under MPI only the first processor with rank==0
         will open the file and write to it. If ``createLocalFiles`` each individual processor will create a file
         where for any processor with rank>0 the file name is extended by its rank. This option is normally only used for
         debug purposes.

         :param fn: filename.
         :type fn: ``str``
         :param append: if True, open file for appending rather than overwriting
         :type append: ``bool``
         :param createLocalFiles: switches on the creation of local files.
         :type createLocalFiles: ``bool``
         """
         error=None
         errno=0
         if len(fn)==0:
             errno=1
             error="No filename provided"
         else:
             self.name=fn
             if append:
                 self.mode='a'
             else:
                 self.mode='w'
             self.__file=None
             self.closed=False
             self.newlines=os.linesep
             # if not the master:
             if getMPIRankWorld()>0:
                  if createLocalFiles:
                      fn2=fn+".%s"%getMPIRankWorld()
                      try:
                         self.__file=open(fn2,self.mode)
                      except Exception as e:
                         errno=1
                         error=e
             else:
                  try:
                      self.__file=open(fn,self.mode)
                  except Exception as e:
                      errno=1
                      error=e
         self.__handelerror(errno, error, "opening")

    def __handelerror(self,errno,e,operation):
         errno=getMPIWorldMax(errno)
         if errno>0:
            if e is None:
               raise IOError("Unable to access file %s in mode %s for %s."%(self.name,self.mode,operation))
            else:
               raise IOError(str(e))

    def close(self):
        """
        Closes the file
        """
        errno=0
        e=None
        try:
           if not self.__file is None:
               self.__file.close()
        except Exception as e:
           errno=1
        self.__handelerror(errno,e,"closing")
        self.closed=True

    def flush(self):
        """
        Flush the internal I/O buffer.
        """
        errno=0
        e=None
        try:
           if not self.__file is None:
               self.__file.flush()
        except Exception as e:
           errno=1
        self.__handelerror(errno,e,"flushing")

    def write(self,txt):
        """
        Write string ``txt`` to file.

        :param txt: string ``txt`` to be written to file
        :type txt: ``str``
        """
        errno=0
        e=None
        try:
           if not self.__file is None:
               self.__file.write(txt)
        except Exception as e:
           errno=1
        self.__handelerror(errno,e,"writing")

    def writelines(self, txts):
        """
        Write the list ``txts`` of strings to the file.

        :param txts: sequence of strings to be written to file
        :type txts: any iterable object producing strings
        :note: Note that newlines are not added. This method is equivalent to calling write() for each string.
        """
        errno=0
        e=None
        try:
           if not self.__file is None:
               self.__file.writelines(txts)
        except Exception as e:
           errno=1
        self.__handelerror(errno,e,"writing strings")




##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################


__copyright__="""Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__="""Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""
__url__="https://github.com/LutzGross/esys-escript.github.io"

"""
Tests that escript returns the correct type when adding real and complex Data objects together

"""

__author__="Adam Ellery, a.ellery@uq.edu.au"

import esys.escriptcore.utestselect as unittest
from esys.escript import *

import numpy

from test_util_base import Test_util_base

class Test_addition_types(Test_util_base):
   """
   these overloaded operations still fail!

        - wrong return value of Data binaries (Mantis 0000054) 
   """
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_add_real_complex_function(self):
      x=Scalar(0, Function(self.domain))
      self.assertTrue(x.isComplex()==False,"Real data object was initiated as complex on Function")
      x+=ComplexScalar(5+8j, Function(self.domain))
      self.assertTrue(x.isComplex(),"Real + Complex is not Complex")

   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_add_complex_real_function(self):
      x=ComplexScalar(5+8j, Function(self.domain))
      self.assertTrue(x.isComplex()==True,"Complex data object was initiated as real on Function")
      x+=Scalar(5, Function(self.domain))
      self.assertTrue(x.isComplex(),"Real + Complex is not Complex")

   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_add_real_complex_continuousfunction(self):
      x=Scalar(0, ContinuousFunction(self.domain))
      self.assertTrue(x.isComplex()==False,"Real data object was initiated as complex on Function")
      x+=ComplexScalar(5+8j, ContinuousFunction(self.domain))
      self.assertTrue(x.isComplex(),"Real + Complex is not Complex")

   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_add_complex_real_continuousfunction(self):
      x=ComplexScalar(5+8j, ContinuousFunction(self.domain))
      self.assertTrue(x.isComplex()==True,"Complex data object was initiated as real on Function")
      x+=Scalar(5, ContinuousFunction(self.domain))
      self.assertTrue(x.isComplex(),"Real + Complex is not Complex")
   
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_add_real_complex_reducedfunction(self):
      x=Scalar(0, ReducedFunction(self.domain))
      self.assertTrue(x.isComplex()==False,"Real data object was initiated as complex on Function")
      x+=ComplexScalar(5+8j, ReducedFunction(self.domain))
      self.assertTrue(x.isComplex(),"Real + Complex is not Complex")

   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_add_complex_real_reducedfunction(self):
      x=ComplexScalar(5+8j, ReducedFunction(self.domain))
      self.assertTrue(x.isComplex()==True,"Complex data object was initiated as real on Function")
      x+=Scalar(5, ReducedFunction(self.domain))
      self.assertTrue(x.isComplex(),"Real + Complex is not Complex")
   


#!/usr/bin/env python3
"""
Convert markdown files to HTML with Sphinx theme
"""

import sys
import os

try:
    import markdown
except ImportError:
    print("ERROR: markdown library not found. Please install python3-markdown")
    sys.exit(1)

def get_sphinx_template(title, content, current_page=""):
    """Generate HTML using Sphinx classic theme structure"""

    # Determine navigation links based on current page
    nav_links = []
    if current_page != "index":
        nav_links.append('<li class="right" style="margin-right: 10px"><a href="index.html">Home</a> |</li>')
    if current_page != "installation":
        nav_links.append('<li class="right"><a href="installation.html">Installation</a> |</li>')
    nav_links.append('<li class="right"><a href="sphinx_api/index.html">Python API</a> |</li>')
    nav_links.append('<li class="right"><a href="user/user.pdf">User Guide (PDF)</a></li>')

    nav_html = '\n        '.join(nav_links)

    template = """<!DOCTYPE html>
<html lang="en" data-content_root="./">
  <head>
    <meta charset="utf-8" />
    <meta name="viewport" content="width=device-width, initial-scale=1.0" />
    <title>{title} &#8212; esys-escript documentation</title>
    <link rel="stylesheet" type="text/css" href="sphinx_api/_static/pygments.css" />
    <link rel="stylesheet" type="text/css" href="sphinx_api/_static/classic.css" />
    <link rel="stylesheet" type="text/css" href="sphinx_api/_static/custom.css" />
  </head>
  <body>
    <div class="related" role="navigation" aria-label="Related">
      <h3>Navigation</h3>
      <ul>
        {nav_links}
        <li class="nav-item nav-item-0"><a href="index.html">esys-escript documentation</a> &#187;</li>
      </ul>
    </div>

    <div class="document">
      <div class="documentwrapper">
        <div class="bodywrapper">
          <div class="body" role="main">
            {content}
            <div class="clearer"></div>
          </div>
        </div>
      </div>
      <div class="sphinxsidebar" role="navigation" aria-label="Main">
        <div class="sphinxsidebarwrapper">
          <h3><a href="index.html">Table of Contents</a></h3>
          <ul>
            <li><a href="index.html">Documentation Home</a></li>
            <li><a href="installation.html">Installation Guide</a></li>
            <li><a href="sphinx_api/index.html">Python API Reference</a></li>
            <li><a href="user/user.pdf">User Guide (PDF)</a></li>
          </ul>
          <h3>Downloads</h3>
          <ul>
            <li><a href="escript_examples.zip">Examples (ZIP)</a></li>
            <li><a href="escript_examples.tar.gz">Examples (TAR.GZ)</a></li>
          </ul>
          <h3>Resources</h3>
          <ul>
            <li><a href="https://github.com/esys-escript/esys-escript.github.io">GitHub Repository</a></li>
            <li><a href="https://github.com/esys-escript/esys-escript.github.io/issues">Report Issues</a></li>
          </ul>
        </div>
      </div>
    </div>

    <div class="footer" role="contentinfo">
      <p>&copy; Copyright 2003-2026, esys.escript group.</p>
      <p>
        Last updated on {date}.
        Created using <a href="https://www.python.org/">Python</a> and
        <a href="https://www.sphinx-doc.org/">Sphinx</a> theme.
      </p>
    </div>
  </body>
</html>
"""

    from datetime import datetime
    date = datetime.now().strftime("%b %d, %Y")

    return template.format(
        title=title,
        content=content,
        nav_links=nav_html,
        date=date
    )

def convert_readme_to_html(md_file, html_file):
    """Convert options.md to index.html with Sphinx theme"""

    with open(md_file, 'r') as f:
        md_content = f.read()

    # Convert markdown to HTML
    html_body = markdown.markdown(md_content, extensions=['extra', 'tables'])

    # Fix links to point to the correct HTML files
    html_body = html_body.replace('href="installation.md"', 'href="installation.html"')
    html_body = html_body.replace('href="./user.pdf"', 'href="user/user.pdf"')

    # Wrap in section tags for Sphinx styling
    html_body = '<section id="esys-escript">\n' + html_body + '\n</section>'

    # Create the documentation resources navigation section
    doc_nav_section = """<section id="documentation-resources">
<h2>Documentation Resources</h2>
<p><strong>Installation Guide</strong></p>
<ul class="simple">
<li><a class="reference external" href="installation.html">Installation Guide (HTML)</a></li>
<li>Source: <code>installation.md</code> in the repository root</li>
</ul>
<p><strong>User Guide (PDF)</strong></p>
<ul class="simple">
<li><a class="reference external" href="user/user.pdf">User Guide PDF</a></li>
</ul>
<p><strong>Python API Reference</strong></p>
<ul class="simple">
<li><a class="reference external" href="sphinx_api/index.html">Python API Documentation</a></li>
</ul>
<p><strong>Examples</strong></p>
<ul class="simple">
<li><a class="reference external" href="escript_examples.zip">Example Scripts (ZIP)</a></li>
<li><a class="reference external" href="escript_examples.tar.gz">Example Scripts (TAR.GZ)</a></li>
</ul>
</section>
"""

    # Insert documentation resources section
    if '<h2>Funding' in html_body or '<h2>The project was funded by the</h2>' in html_body:
        html_body = html_body.replace('<h2>Funding', doc_nav_section + '<h2>Funding')
        html_body = html_body.replace('<h2>The project was funded by the</h2>',
                                     doc_nav_section + '<h2>The project was funded by the</h2>')
    else:
        # Fallback: add before closing section tag
        html_body = html_body.replace('</section>', doc_nav_section + '</section>')

    full_html = get_sphinx_template("esys-escript", html_body, "index")

    with open(html_file, 'w') as f:
        f.write(full_html)

    print(f"Converted {md_file} to {html_file}")

def convert_installation_to_html(md_file, html_file):
    """Convert installation.md to HTML with Sphinx theme"""

    with open(md_file, 'r') as f:
        md_content = f.read()

    # Convert markdown to HTML
    html_body = markdown.markdown(md_content, extensions=['extra', 'tables', 'fenced_code', 'toc'])

    # Fix links to point to the correct HTML files
    html_body = html_body.replace('href="./scons/templates/options.md"', 'href="options.html"')

    # Wrap in section tag for Sphinx styling
    html_body = '<section id="installation-guide">\n' + html_body + '\n</section>'

    full_html = get_sphinx_template("Installation Guide", html_body, "installation")

    with open(html_file, 'w') as f:
        f.write(full_html)

    print(f"Converted {md_file} to {html_file}")

def convert_options_to_html(md_file, html_file):
    """Convert options.md to HTML with Sphinx theme"""

    with open(md_file, 'r') as f:
        md_content = f.read()

    # Convert markdown to HTML
    html_body = markdown.markdown(md_content, extensions=['extra', 'tables'])

    full_html = get_sphinx_template("Build Options Reference", html_body, "options")

    with open(html_file, 'w') as f:
        f.write(full_html)

    print(f"Converted {md_file} to {html_file}")

if __name__ == '__main__':
    if len(sys.argv) < 4:
        print("Usage: md2html.py <type> <input.md> <output.html>")
        print("  type: 'readme', 'installation', or 'options'")
        sys.exit(1)

    doc_type = sys.argv[1]
    input_file = sys.argv[2]
    output_file = sys.argv[3]

    if not os.path.exists(input_file):
        print(f"ERROR: Input file {input_file} does not exist")
        sys.exit(1)

    # Ensure output directory exists
    output_dir = os.path.dirname(output_file)
    if output_dir and not os.path.exists(output_dir):
        os.makedirs(output_dir)

    if doc_type == 'readme':
        convert_readme_to_html(input_file, output_file)
    elif doc_type == 'installation':
        convert_installation_to_html(input_file, output_file)
    elif doc_type == 'options':
        convert_options_to_html(input_file, output_file)
    else:
        print(f"ERROR: Unknown type '{doc_type}'. Use 'readme', 'installation', or 'options'")
        sys.exit(1)




##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################

# This is a template configuration file for escript on macOS with Homebrew.
# Refer to installation.md for usage instructions.
#
# Prerequisites:
#   brew install python3 scons cmake llvm
#   brew install boost boost-python3
#   brew install hdf5 suite-sparse netcdf netcdf-cxx4 zlib metis openblas scotch
#   brew install open-mpi mpi4py
#   brew install numpy scipy python-matplotlib
#
# Optional (sympy not available in Homebrew):
#   For symbolic math support, install sympy in a virtual environment:
#   python3 -m venv ~/venv && source ~/venv/bin/activate && pip install sympy
#
# For LLVM compiler support (recommended for OpenMP):
#   echo 'export PATH="/opt/homebrew/opt/llvm/bin:$PATH"' >> ~/.zshrc
#   export LDFLAGS="-L/opt/homebrew/opt/llvm/lib"
#   export CPPFLAGS="-I/opt/homebrew/opt/llvm/include"
#
# For no-MPI build, use homebrew_nompi_options.py instead.

import os
import subprocess

escript_opts_version = 203

# Homebrew prefix (Apple Silicon: /opt/homebrew, Intel: /usr/local)
HOMEBREW_PREFIX = '/opt/homebrew'

# Python configuration - auto-detect version
pythoncmd = os.path.join(HOMEBREW_PREFIX, 'bin', 'python3')
p = subprocess.run([pythoncmd, '-V'], capture_output=True, text=True)
print(p.stdout)
subversion = p.stdout.split(' ')[1].split('.')[1].strip()
revversion = p.stdout.split(' ')[1].split('.')[2].strip()
print(f'Compiled for Python 3.{subversion}.{revversion}.')

# Get Python include and library paths
p = subprocess.run([pythoncmd + f".{subversion}-config", '--includes'], capture_output=True, text=True)
for t in p.stdout.split(" "):
    if t.startswith("-I"):
        pythonincpath = t[2:]
        break

p = subprocess.run([pythoncmd + f".{subversion}-config", '--ldflags'], capture_output=True, text=True)
for t in p.stdout.split(" "):
    if t.startswith("-L"):
        pythonlibpath = t[2:]
        break

# Set SDKROOT for macOS builds
try:
    p = subprocess.run(['xcrun', '--show-sdk-path'], capture_output=True, text=True)
    if p.stdout.strip():
        os.putenv("SDKROOT", p.stdout.strip())
except FileNotFoundError:
    print("xcrun not found. SDKROOT is not set.")

# Compiler configuration - use LLVM for OpenMP support
# Install with: brew install llvm
cxx = '/opt/homebrew/bin/g++-15'
cc = '/opt/homebrew/bin/gcc-15'

cxx_extra = ["-Wno-error=uninitialized",
             "-Wno-error=maybe-uninitialized",
             "-Wno-error=implicit-function-declaration",
             "-Wno-error=unused-but-set-variable",
             "-Wno-error=return-local-addr",
             "",
             "-Wno-error=unused-function",
             ""]  # Allow C99 _Complex in LAPACK headers

cc_extra = ["-Wno-error=unused-but-set-variable",
            "-Wno-error=deprecated-non-prototype"]
cc_optim = ["-O3"]

print(f"C++ compiler is {cxx}")

# Skip library link checks on macOS
# The SCons configuration-time link tests can fail on macOS with Python 3.14+
# due to environment issues, but actual linking works fine during build.
# This only affects configuration checks, not the actual build linking.
skip_link_checks = True

# OpenMP configuration
openmp = True
omp_flags = ["-fopenmp"]
omp_ldflags = ["-fopenmp"]

# MPI configuration
# Install with: brew install open-mpi mpi4py
mpi = 'none'
# OpenMPI 5.x uses the standard include/lib directories (not openmpi subdirectories)
mpi_prefix = [HOMEBREW_PREFIX + '/include', HOMEBREW_PREFIX + '/lib']
# OpenMPI 5.x dropped libmpi_cxx - C++ bindings are now in libmpi
mpi_libs = ['mpi']
mpi4py = False

# Solver configuration
paso = True
build_trilinos = 'check'

# Domain configuration - oxley excluded (requires p4est)
domains = ('finley', 'ripley', 'speckley')

# Boost configuration
boost_prefix = HOMEBREW_PREFIX
boost_libs = [f'boost_python3{subversion}']
compression_libs = ['boost_iostreams']

# HDF5 configuration
hdf5 = True
hdf5_prefix = HOMEBREW_PREFIX
hdf5_libs = ['hdf5_cpp', 'hdf5']

# SILO configuration (for VisIt visualization)
silo = True
silo_prefix = HOMEBREW_PREFIX
silo_libs = ['siloh5', 'hdf5']

# NetCDF configuration
netcdf = True
netcdf_prefix = HOMEBREW_PREFIX
netcdf_libs = ['netcdf-cxx4', 'netcdf']

# UMFPACK direct solver
umfpack = True
umfpack_prefix = [HOMEBREW_PREFIX + "/include/suitesparse", HOMEBREW_PREFIX + "/lib/"]

# LAPACK configuration - use LAPACKE from OpenBLAS (keg-only)
# LAPACKE is the C interface to LAPACK, compatible with C++
# Install with: brew install openblas
lapack = True
lapack_prefix = [HOMEBREW_PREFIX + '/opt/openblas/include', HOMEBREW_PREFIX + '/opt/openblas/lib']
lapack_libs = ['openblas']  # OpenBLAS includes LAPACKE interface

# zlib - keg-only in Homebrew (macOS provides system version)
# Use Homebrew version for consistency: brew install zlib
zlib = True
zlib_prefix = [HOMEBREW_PREFIX + '/opt/zlib/include', HOMEBREW_PREFIX + '/opt/zlib/lib']
zlib_libs = ['z']

# METIS graph partitioning library (used by Trilinos)
# Install with: brew install metis
metis = True
metis_prefix = [HOMEBREW_PREFIX + '/include', HOMEBREW_PREFIX + '/lib']
metis_libs = ['metis']

# ParMETIS - not available in Homebrew, would need to build from source
parmetis = False

# Scotch/PT-Scotch graph partitioning library (alternative to ParMETIS)
# Install with: brew install scotch
# PT-Scotch provides parallel graph partitioning like ParMETIS
scotch = True
scotch_prefix = [HOMEBREW_PREFIX + '/opt/scotch/include', HOMEBREW_PREFIX + '/opt/scotch/lib']
scotch_libs = ['ptscotch', 'ptscotcherr', 'scotch', 'scotcherr']

# Linker flags - include LLVM libc++ for macOS arm64
ld_extra = ["-lz"]  # GCC uses libstdc++ instead of libc++

# Optional features
# sympy not available in Homebrew - requires virtual environment installation
sympy = False

# Clean up temporary variables
del HOMEBREW_PREFIX, p, subversion, revversion, t



##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################

"""General cost functions for minimization"""


__copyright__ = """Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__ = """Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""
__url__ = "https://github.com/LutzGross/esys-escript.github.io"

__all__ = ['CostFunction']

class CostFunction(object):
    """
    A cost function *F(m)* that can be minimized (base class).
    Near a solution *m* in the search space the cost function is approximated 
    in direction *p* as       

    *F(m+p) ~ F(m) + <grad F(m), p> + < H p, p>*
    
    where *F(m)* is the value of the function at *m*, *grad F* is the gradient, *H* is the Hessian and
    *<.,.>* is dual product. These four procedures are defined in this class.

    The class distinguishes between the representation of the solution
    m (m-type) and the cost function gradient (g-type).

    Example of usage::

        class DerivedCostFunction(CostFunction):
            # overwrite: getArguments, etc.
            pass

        cf = DerivedCostFunction()
        # ... calculate m ...
        args = cf.getArguments(m)  # this could be potentially expensive!
        f = cf.getValue(m, *args)
        # ... it could be required to update m without using the gradient...
        # ... but then ...
        gf = cf.getGradient(m, *args)

    Use the "AndCount" versions (e.g. getValueAndCount) if you want to count calls.

    """
    def __init__(self):
        """
        Constructor. Initializes logger.
        """
        #self.logger = logging.getLogger('esys.%s' % self.__class__.__name__)
        self.resetStatistics()

    def resetStatistics(self):
        """
        resets all counters
        """
        self.DualProduct_calls = 0
        self.Value_calls = 0
        self.Gradient_calls = 0
        self.Arguments_calls = 0
        self.InverseHessianApproximation_calls = 0
        self.Norm_calls = 0
        
    def getStatistics(self):
        """
        Returns the call statistics as a string.

        :return: formatted string with counts of all function calls
        :rtype: ``str``
        """
        out="Number of cost function evaluations: %d\n" % self.Value_calls
        out+="Number of gradient evaluations: %d\n" % self.Gradient_calls
        out+="Number of inverse Hessian evaluations: %d\n" % self.InverseHessianApproximation_calls
        out+="Number of inner product evaluations: %d\n" % self.DualProduct_calls
        out+="Number of argument evaluations: %d\n" % self.Arguments_calls
        out+="Number of norm evaluations: %d" % self.Norm_calls
        return out

    def __call__(self, m, *args):
        """
        short for `sum(getValueAndCount(m, *args))`.
        """
        return self.getValueAndCount(m, *args)
    
    def getDualProductAndCount(self, m, r):
        """
        returns the dual product *<.,.>* of ``r`` and ``m``.

        When calling this method the calling statistics is updated.

        :type m: m-type
        :type r: g-type
        :rtype: ``float``
        """
        self.DualProduct_calls += 1
        return self.getDualProduct(m, r)

    def getNormAndCount(self, m):
        """
        returns the norm of ``m``.

        When calling this method the calling statistics is updated.

        :type m: m-type
        :rtype: ``float``
        """
        self.Norm_calls += 1
        return self.getNorm(m)

    def getArgumentsAndCount(self, m):
        """
        returns precalculated values that are shared in the calculation of
        *F(m)* and *grad F(m)* and the Hessian operator. The default
        implementation returns an empty tuple.

        When calling this method the calling statistics is updated.

        :param m: location of derivative
        :type m: m-type
        :rtype: ``tuple``

        :note: The tuple returned by this call will be passed back to this `CostFunction` in other
           calls (eg: `getGradient`). Its contents are not specified at this level because no code,
           other than the `CostFunction` which created it, will be interacting with it.
           That is, the implementor can put whatever information they find useful in it.

        """
        self.Arguments_calls += 1
        return self.getArguments(m)

    def getValueAndCount(self, m, *args):
        """
        returns the value *F(m)* using the precalculated values for *m*.

        When calling this method the calling statistics is updated.

        :param m: a solution approximation
        :type m: m-type
        :param args: pre-calculated values for ``m`` from `getArgumentsAndCount()`
        :rtype: ``float`` or valid argument for ``sum``
        """
        self.Value_calls += 1
        if not args:
            args = self.getArgumentsAndCount(m)
        return self.getValue(m, *args)

    def getGradientAndCount(self, m, *args):
        """
        returns the gradient of *F* at *m* using the precalculated values for *m*.

        When calling this method the calling statistics is updated.

        :param m: location of derivative
        :type m: m-type
        :param args: pre-calculated values for ``m`` from `getArgumentsAndCount()`
        :rtype: g-type
        """
        self.Gradient_calls += 1
        if not args:
            args = self.getArgumentsAndCount(m)
        return self.getGradient(m, *args)

    def getInverseHessianApproximationAndCount(self, r, m,  *args, initializeHessian=True):
        """
        returns an evaluation *p* of the inverse of the Hessian
        operator of the cost function for a given gradient *r*: *H p = r*

        When calling this method the calling statistics is updated.

        :param r: a given gradient
        :type r: g-type
        :param initializeHessian: indicates if the Hessian operator should be initialized using `m`.
                                    If updating the Hessian is expensive it should only be done
                                    when initializeHessian is True. If this method provides an approximation
                                    only and building the new Hessian approximation is expensive
                                    it is typically more efficient to update the Hessian operator
                                    occasionally.
        :type initializeHessian: bool
        :param args: pre-calculated values for ``m`` from `getArgumentsAndCount()`
        :returns: new search direction p.
        :rtype: m-type
        """
        if not args:
            args = self.getArgumentsAndCount(m)
        self.InverseHessianApproximation_calls += 1
        return self.getInverseHessianApproximation(r, m,  *args, initializeHessian=initializeHessian)

    def getDualProduct(self, m, r):
        """
        returns the dual product of ``m`` and ``r``

        :type m: m-type
        :type r: g-type
        :rtype: ``float``
        :note: Overwrite this method to implement a cost function.
        """
        raise NotImplementedError

    def getNorm(self, m):
        """
        returns the norm of ``m``.
        Typically, this is the 'Lsup' function.

        :type m: m-type
        :rtype: ``float``
        :note: Overwrite this method to implement a cost function.
        """
        raise NotImplementedError

    def getArguments(self, m):
        """
        returns precalculated values that are shared in the calculation of
        *F(m)* and *grad F(m)* and the Hessian operator.

        :note: Overwrite this method to implement a cost function.
        :note: The tuple returned by this call will be passed back to this `CostFunction` in other
           calls(eg: ``getGradientAndCount``). Its contents are not specified at this level because no code,
           other than the `CostFunction` which created it, will be interacting with it.
           That is, the implementor can put whatever information they find useful in it.

        :param m: location of derivative
        :type m: m-type
        :returns: pre-calculated arguments
        :rtype: ``tuple``

        :note: Overwrite this method to implement a cost function.
        """
        return ()

    def getValue(self, m, *args):
        """
        returns the value *F(m)* using the precalculated values for *m*.

        :param m: a solution approximation
        :type m: m-type
        :param args: pre-calculated values for ``m`` from `getArgumentsAndCount()`
        :rtype: ``float`` or valid argument for ``sum``

        :note: Overwrite this method to implement a cost function.
        """
        raise NotImplementedError

    def getGradient(self, m, *args):
        """
        returns the gradient of *F* at *m* using the precalculated values for
        *m*.

        :param m: location of derivative
        :type m: m-type
        :param args: pre-calculated values for ``m`` from `getArgumentsAndCount()`
        :rtype: g-type
        :note: Overwrite this method to implement a cost function.
        """
        raise NotImplementedError

    def getSqueezeFactor(self, m, p, *args):
        """
        The new solution is calculated as m+a*p with a>0. This function allows to provide an upper bound
        for a to make sure that m+a*p is valid typically to avoid overflow when the cost function is evaluated.
        the solver will take action to make sure that the value of a is not too small.

        :param m: a solution approximation
        :type m: m-type
        :param p: an increment to the solution
        :type m: m-type
        :param args: pre-calculated values for ``m`` from `getArgumentsAndCount()`
        :rtype: positive ``float`` or None

        :note: Overwrite this method to implement a cost function.
        """
        return None

    def getInverseHessianApproximation(self, r, m, *args, initializeHessian = True):
        """
        returns an approximate evaluation *p* of the inverse of the Hessian
        operator of the cost function for a given gradient *r*: *H p = r*
        :note: by default this method is returning *r*. In this case it is assumed that m-type==g-type
        :note: Overwrite this method to implement a cost function.

        :param r: a given gradient
        :type r: g-type
        :param initializeHessian: indicates if the Hessian operator should be initialized using `m`.
                                    If this method provides an approximation
                                    only and building the new Hessian approximation is expensive
                                    it is typically more efficient to update the Hessian operator
                                    occasionally only when on input initializeHessian is `True`.
                                    If the Hessian should be updated in each step ignore the
                                    value of `initializeHessian`.
        :type initializeHessian: bool
        :returns: new search direction p.
        :rtype: m-type

        """
        return r


##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################


"""
Provides :class:`InterpolationTable`, a class-based interface for
interpolating values from a regular-grid lookup table onto a mesh.
Supports order-0 (piecewise constant) and order-1 (linear) interpolation
in 1D, 2D, and 3D.  The lookup table may contain real or complex values;
the coordinate data *x* must always be real.
"""

import numpy as np
from . import escriptcpp as escore


class InterpolationTable:
    """
    Interpolates values from a regular-grid lookup table onto a mesh.

    The coordinate data *x* passed to :meth:`__call__` determines the
    interpolation dimension:

    ========================  ===========  =======================
    ``x.getShape()``          Lookup dim   Required table rank
    ========================  ===========  =======================
    ``()``                    1-D          1  — shape ``(nx,)``
    ``(1,)``                  1-D          1  — shape ``(nx,)``
    ``(2,)``                  2-D          2
    ``(3,)``                  3-D          3
    ========================  ===========  =======================

    The returned `Data` object is always **scalar** (shape ``()``).
    If the table contains complex values the result is complex `Data`;
    the coordinate `Data` *x* must always be real.

    **Table indexing convention** (controlled by *table_indexing*):

    * ``"ijk"`` *(default)* — natural numpy order: the first index varies
      along the x-axis, the second along y, the third along z.
      Shapes: ``(nx,)`` / ``(nx, ny)`` / ``(nx, ny, nz)``.

    * ``"kji"`` — legacy / C++ order: the first index varies along the
      z-axis, the second along y, the third along x.
      Shapes: ``(nx,)`` / ``(ny, nx)`` / ``(nz, ny, nx)``.

    Example usage::

        import numpy as np
        from esys.finley import Rectangle
        from esys.escript import Function
        from esys.escriptcore.interpolation import InterpolationTable

        dom = Rectangle(20, 20)
        x = Function(dom).getX()   # shape (2,)

        # 2-D scalar table, "ijk" convention: shape (nx, ny)
        t = np.random.rand(5, 5)
        interp = InterpolationTable(t, origin=(0., 0.), step=(0.25, 0.25))
        result = interp(x)

        # Equivalent using total extent (covers [0,1] x [0,1])
        interp = InterpolationTable(t, origin=(0., 0.), extend=(1., 1.))
        result = interp(x)

        # Order-0 (piecewise constant) interpolation
        interp0 = InterpolationTable(t, origin=(0., 0.), extend=(1., 1.), order=0)
        result0 = interp0(x)

    :param table: lookup table as a numpy array of rank 1, 2, or 3
    :type table: ``numpy.ndarray``
    :param origin: coordinate(s) of the first table entry; a single float
        for 1-D or a tuple for 2-D / 3-D
    :type origin: ``float`` or ``tuple`` of ``float``
    :param step: cell size(s), i.e. the spacing between adjacent table
        entries; all values must be strictly positive.  Exactly one of
        *step* and *extend* must be supplied.
    :type step: ``float`` or ``tuple`` of ``float``
    :param extend: total extent of the table domain along each coordinate
        axis.  For order-1: distance from the first to the last sample
        point along axis *i* (``step[i] = extend[i] / (n_i - 1)``).
        For order-0: total domain width along axis *i*
        (``step[i] = extend[i] / n_i``), where *n_i* is the number of
        cells along that axis.  Exactly one of *step* and *extend* must
        be supplied.
    :type extend: ``float`` or ``tuple`` of ``float``
    :param order: interpolation order — 0 for piecewise constant, 1 for
        linear (default)
    :type order: ``int``
    :param undef: upper threshold; result values above this trigger a
        ``RuntimeError``
    :type undef: ``float``
    :param check_boundaries: if ``True``, a ``RuntimeError`` is raised when
        a coordinate lies outside the table extent; otherwise the nearest
        boundary value is used
    :type check_boundaries: ``bool``
    :param table_indexing: indexing convention for the table array.
        ``"ijk"`` (default) — first index = x-axis, natural numpy order.
        ``"kji"`` — first index = z-axis, legacy C++ order.
    :type table_indexing: ``str``
    """

    def __init__(self, table, origin, step=None, extend=None, order=1,
                 undef=1.e50, check_boundaries=False, table_indexing="ijk"):
        if not isinstance(table, np.ndarray):
            table = np.asarray(table)
        # Promote integer tables to float; preserve complex dtype
        if np.issubdtype(table.dtype, np.integer):
            table = table.astype(float)
        if np.isscalar(origin):
            origin = (origin,)
        ndim = len(origin)
        if ndim < 1 or ndim > 3:
            raise ValueError("ndim (length of origin) must be 1, 2, or 3")
        if table.ndim != ndim:
            raise ValueError(
                "table rank {} does not match coordinate dimension {} "
                "(set by length of origin)".format(table.ndim, ndim))
        if order not in (0, 1):
            raise ValueError("order must be 0 or 1")
        if table_indexing not in ("ijk", "kji"):
            raise ValueError("table_indexing must be 'ijk' or 'kji'")

        # n[i] = number of table entries along coordinate axis i (0=x,1=y,2=z)
        if table_indexing == "ijk":
            self._n = tuple(int(table.shape[i]) for i in range(ndim))
            # Transpose to internal "kji" storage for C++ workers.
            # 1-D: no change; 2-D: axes (0,1)->(1,0); 3-D: (0,1,2)->(2,1,0)
            if ndim == 1:
                self._table = np.ascontiguousarray(table)
            elif ndim == 2:
                self._table = np.ascontiguousarray(table.T)
            else:
                self._table = np.ascontiguousarray(table.transpose(2, 1, 0))
        else:  # "kji"
            self._n = tuple(int(table.shape[ndim - 1 - i]) for i in range(ndim))
            self._table = np.ascontiguousarray(table)
        self._is_complex = np.iscomplexobj(self._table)

        # Resolve step / extend using n[i] (entries along axis i)
        if step is None and extend is None:
            raise ValueError("Exactly one of 'step' or 'extend' must be provided")
        if step is not None and extend is not None:
            raise ValueError("Only one of 'step' or 'extend' may be provided, not both")
        if extend is not None:
            if np.isscalar(extend):
                extend = (extend,)
            if len(extend) != ndim:
                raise ValueError("extend must have the same length as origin")
            if order == 0:
                # n cells each of width step; domain is [origin, origin+extend]
                step = tuple(float(e) / self._n[i]
                             for i, e in enumerate(extend))
            else:
                # n-1 intervals between n sample points
                step = tuple(float(e) / (self._n[i] - 1)
                             for i, e in enumerate(extend))
        else:
            if np.isscalar(step):
                step = (step,)
            if len(step) != ndim:
                raise ValueError("step must have the same length as origin")
            step = tuple(float(s) for s in step)

        if any(s <= 0 for s in step):
            raise ValueError("All step sizes must be strictly positive")
        self._origin = tuple(float(v) for v in origin)
        self._step = step
        self._ndim = ndim
        self._order = order
        self._undef = float(undef)
        self._check_boundaries = check_boundaries
        self._table_indexing = table_indexing

    # ------------------------------------------------------------------
    # Internal C++ dispatch (table must be a real contiguous numpy array)
    # ------------------------------------------------------------------

    def _cpp_1d(self, x0, table):
        if self._order == 0:
            return x0._interpolateTable1dOrder0(
                table, self._origin[0], self._step[0],
                self._undef, self._check_boundaries)
        else:
            return x0.interpolateTable(
                table, self._origin[0], self._step[0],
                self._undef, self._check_boundaries)

    def _cpp_2d(self, x0, x1, table):
        if self._order == 0:
            return x0._interpolateTable2dOrder0(
                table, self._origin[0], self._step[0],
                x1, self._origin[1], self._step[1],
                self._undef, self._check_boundaries)
        else:
            return x0.interpolateTable(
                table, self._origin[0], self._step[0],
                x1, self._origin[1], self._step[1],
                self._undef, self._check_boundaries)

    def _cpp_3d(self, x0, x1, x2, table):
        if self._order == 0:
            return x0._interpolateTable3dOrder0(
                table, self._origin[0], self._step[0],
                x1, self._origin[1], self._step[1],
                x2, self._origin[2], self._step[2],
                self._undef, self._check_boundaries)
        else:
            return x0._interpolateTable3d(
                table, self._origin[0], self._step[0],
                x1, self._origin[1], self._step[1],
                x2, self._origin[2], self._step[2],
                self._undef, self._check_boundaries)

    def _do_call(self, table, x):
        """Dispatch to the correct C++ worker using a real numpy *table*."""
        sh = x.getShape()
        if sh == ():
            return self._cpp_1d(x, table)
        if sh == (1,):
            return self._cpp_1d(x[0], table)
        if sh == (2,):
            return self._cpp_2d(x[0], x[1], table)
        # sh == (3,)
        return self._cpp_3d(x[0], x[1], x[2], table)

    # ------------------------------------------------------------------
    # Public interface
    # ------------------------------------------------------------------

    def __call__(self, x):
        """
        Return interpolated values at the mesh points described by *x*.

        :param x: coordinate data with shape ``()``, ``(1,)``, ``(2,)``, or
            ``(3,)``; the shape must be consistent with the table rank.
            *x* must be real — complex coordinate data is not supported.
        :type x: `Data`
        :return: interpolated field (complex if the table contains complex
            values, otherwise real)
        :rtype: `Data` with shape ``()``
        :raises TypeError: if *x* is not a real `Data` object
        :raises ValueError: if *x* has an unsupported shape or is inconsistent
            with the table rank
        """
        if not isinstance(x, escore.Data):
            raise TypeError("x must be a Data object")
        if x.isComplex():
            raise TypeError(
                "coordinate data x must be real; complex x is not supported")

        sh = x.getShape()
        expected_ndim = {(): 1, (1,): 1, (2,): 2, (3,): 3}
        if sh not in expected_ndim:
            raise ValueError(
                "x must have shape (), (1,), (2,), or (3,); got {}".format(sh))
        if expected_ndim[sh] != self._ndim:
            raise ValueError(
                "x with shape {} requires ndim={}, got ndim={}".format(
                    sh, expected_ndim[sh], self._ndim))

        if self._is_complex:
            t_re = np.ascontiguousarray(self._table.real)
            t_im = np.ascontiguousarray(self._table.imag)
            return self._do_call(t_re, x) + 1j * self._do_call(t_im, x)
        return self._do_call(self._table, x)

    def interpolate(self, x):
        """Alias for :meth:`__call__`."""
        return self(x)

    # ------------------------------------------------------------------
    # Read-only accessors
    # ------------------------------------------------------------------

    def getOrder(self):
        """Return the interpolation order (0 or 1)."""
        return self._order

    def getNdim(self):
        """Return the number of coordinate dimensions (1, 2, or 3)."""
        return self._ndim

    def getOrigin(self):
        """Return the tuple of starting coordinates."""
        return self._origin

    def getStep(self):
        """Return the tuple of cell sizes."""
        return self._step

    def getTableIndexing(self):
        """Return the table indexing convention (``'ijk'`` or ``'kji'``)."""
        return self._table_indexing

    def getExtend(self):
        """Return the tuple of domain extents along each coordinate axis.

        For order-1: ``extend[i] = step[i] * (n_i - 1)``
        (distance from first to last sample point along axis *i*).

        For order-0: ``extend[i] = step[i] * n_i``
        (total domain width; domain along axis *i* is
        ``[origin[i], origin[i] + extend[i]]``).

        Here *n_i* is the number of table entries along coordinate axis *i*
        (independent of the *table_indexing* convention).
        """
        if self._order == 0:
            return tuple(s * self._n[i] for i, s in enumerate(self._step))
        else:
            return tuple(s * (self._n[i] - 1) for i, s in enumerate(self._step))




##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################

"""
Domain Array and Data Coupler for Multi-Domain Simulations
============================================================

This module provides infrastructure for multi-domain parallel simulations
using MPI communicator topology and data exchange between domains.

Classes
-------

- **MPIDomainArray**: MPI communicator topology for domain arrays.
  Manages 2D Cartesian communicator topology with domain and subdomain
  communicators for multi-domain simulations.

- **DataCoupler**: Communication of Data objects between domains.
  Handles sending, receiving, broadcasting, and reduction of esys.escript.Data
  objects between identically distributed domains.

Example Usage
-------------

::

    from mpi4py import MPI
    from esys.escript import Rectangle, Scalar, Solution
    from esys.escript.domaincoupler import MPIDomainArray, DataCoupler

    # Create domain array with 2 domains
    domain_array = MPIDomainArray(numDomains=2, comm=MPI.COMM_WORLD)

    # Create domains (same mesh parameters for identical distribution)
    domain = Rectangle(n0=50, n1=50, comm=domain_array.getDomainComm())

    # Create coupler for data exchange
    coupler = DataCoupler(domain_array)

    # Exchange data between domains
    if domain_array.getDomainIndex() == 0:
        data = Scalar(1.0, Solution(domain))
        coupler.send(data, dest_domain_index=1, tag=100)
    elif domain_array.getDomainIndex() == 1:
        received = coupler.receive(Solution(domain), source_domain_index=0, tag=100)

"""

from mpi4py import MPI
import numpy as np
from esys.escript import Data


__all__ = ['MPIDomainArray', 'DataCoupler']


class MPIDomainArray(object):
    """
    MPI communicator topology for multi-domain simulations.

    Creates a 2D Cartesian communicator topology where:
    - First dimension: different domains (e.g., different physics, ensemble members)
    - Second dimension: subdomains within each domain (for domain decomposition)

    This allows each domain to have its own MPI communicator for parallel operations
    while enabling communication between corresponding ranks across domains.

    Parameters:
    -----------
    numDomains : int
        Number of independent domains
    numSubDomains : int, optional
        Number of MPI ranks per domain. If None, computed as total_ranks // numDomains
    strict : bool, optional
        If True, requires total ranks to exactly match numDomains * numSubDomains.
        Default: True
    comm : MPI.Comm, optional
        Parent MPI communicator. Default: MPI.COMM_WORLD

    Attributes:
    -----------
    numDomains : int
        Number of domains
    numSubDomains : int
        Number of subdomains per domain
    comm : MPI.Comm
        Parent communicator
    domain_comm : MPI.Comm
        Communicator for ranks within same domain (for intra-domain parallelization)
    subdomain_comm : MPI.Comm
        Communicator for corresponding ranks across domains (for inter-domain communication)
    idom : int
        This rank's domain index (0 to numDomains-1)
    isub : int
        This rank's subdomain index (0 to numSubDomains-1)

    Example::

        # 8 total ranks: 2 domains x 4 subdomains each
        domain_array = MPIDomainArray(numDomains=2, comm=MPI.COMM_WORLD)

        # Create domain using the domain communicator
        from esys.ripley import Rectangle
        domain = Rectangle(n0=100, n1=100, comm=domain_array.getDomainComm())

        # Ranks 0-3: domain 0, subdomains 0-3
        # Ranks 4-7: domain 1, subdomains 0-3
        # subdomain_comm connects rank 0 with rank 4, rank 1 with rank 5, etc.
    """

    def __init__(self, numDomains, numSubDomains=None, strict=True, comm=None):
        self.numDomains = numDomains
        if comm is None:
            self.comm = MPI.COMM_WORLD
        else:
            self.comm = comm

        if numSubDomains is None:
            numSubDomains = self.comm.Get_size() // self.numDomains
        self.numSubDomains = numSubDomains

        myrank = self.comm.Get_rank()
        size = self.comm.Get_size()

        if strict and not numDomains * numSubDomains == size:
            raise ValueError(f"Requested ranks {numDomains * numSubDomains} does not match available ranks {size}")
        if numDomains * numSubDomains > size:
            raise ValueError(f"Requested ranks {numDomains*numSubDomains} is larger than available ranks {size}")

        cc = comm.Create_cart([numDomains, numSubDomains], periods=[False, True], reorder=False)
        if myrank < cc.Get_size():
           self.group_by_subdomain = {}
           for isub in range(cc.dims[1]):
               dom_group_list = [cc.Get_cart_rank((idom, isub)) for idom in range(cc.dims[0])]
               if myrank in dom_group_list:
                   self.isub = isub
                   self.dom_group_list = dom_group_list
               self.group_by_subdomain[isub] = comm.Get_group().Incl(dom_group_list)
           self.subdomain_comm = self.comm.Create(self.group_by_subdomain[self.isub])

           self.group_by_domain = {}
           for idom in range(cc.dims[0]):
                    sub_group_list = [cc.Get_cart_rank((idom, isub)) for isub in range(cc.dims[1])]
                    if myrank in sub_group_list:
                        self.idom = idom
                        self.sub_group_list = sub_group_list
                    self.group_by_domain[idom] = comm.Get_group().Incl(sub_group_list)
           self.domain_comm = self.comm.Create(self.group_by_domain[self.idom])

    def getSubdomainComm(self):
        """
        Get the subdomain communicator.

        Returns:
        --------
        MPI.Comm
            Communicator connecting corresponding ranks across all domains.
            Size equals numDomains.
        """
        return self.subdomain_comm

    def getDomainComm(self):
        """
        Get the domain communicator.

        Returns:
        --------
        MPI.Comm
            Communicator for all ranks within this domain.
            Size equals numSubDomains. Use this for domain creation.
        """
        return self.domain_comm

    def getSubdomainIndex(self):
        """
        Get this rank's subdomain index within its domain.

        Returns:
        --------
        int
            Subdomain index (0 to numSubDomains-1)
        """
        return self.isub

    def getDomainIndex(self):
        """
        Get this rank's domain index.

        Returns:
        --------
        int
            Domain index (0 to numDomains-1)
        """
        return self.idom


class DataCoupler:
    """
    Handles communication of Data objects between identically distributed domains.

    This class manages point-to-point and collective communication of esys.escript.Data
    objects between domains in an MPIDomainArray. Each rank communicates with its
    corresponding rank in other domains using the subdomain communicator.

    Key Assumption:
        Source and target domains must have the same FunctionSpace structure and
        identical domain decomposition (same number of samples per rank, same distribution).

        This is guaranteed when using MPIDomainArray where all domains are created with
        the same mesh parameters and distributed across communicators of the same size.

    Parameters:
    -----------
    mpi_domain_array : MPIDomainArray
        The domain array managing communicator topology

    Methods:
    --------
    send(data, dest_domain_index, tag=0)
        Send Data to corresponding rank in destination domain

    receive(function_space, source_domain_index, tag=0)
        Receive Data from corresponding rank in source domain

    exchange(send_data, recv_function_space, peer_domain_index, tag=0)
        Bidirectional exchange between two domains

    broadcast(data=None, function_space=None, root_domain_index=0)
        Broadcast Data from root domain to all domains

    allreduce(data, op=MPI.SUM)
        All-reduce Data point-by-point across all domains

    broadcast_value(value=None, root_domain_index=0)
        Broadcast scalar/array from root domain

    allreduce_value(value, op=MPI.SUM)
        All-reduce scalar/array across all domains

    Example::

        from mpi4py import MPI
        from esys.ripley import Rectangle
        from esys.escript import Scalar, Solution
        from esys.escript.domaincoupler import MPIDomainArray, DataCoupler

        # Setup domain array
        domain_array = MPIDomainArray(numDomains=2, comm=MPI.COMM_WORLD)
        domain = Rectangle(n0=50, n1=50, comm=domain_array.getDomainComm())
        coupler = DataCoupler(domain_array)

        # Point-to-point communication
        if domain_array.getDomainIndex() == 0:
            data = Scalar(1.0, Solution(domain))
            coupler.send(data, dest_domain_index=1, tag=100)
        elif domain_array.getDomainIndex() == 1:
            received = coupler.receive(Solution(domain), source_domain_index=0, tag=100)

        # Collective operations
        my_data = Scalar(domain_array.getDomainIndex(), Solution(domain))
        sum_data = coupler.allreduce(my_data, op=MPI.SUM)
    """

    def __init__(self, mpi_domain_array):
        """
        Initialize DataCoupler with an MPIDomainArray.

        Parameters:
        -----------
        mpi_domain_array : MPIDomainArray
            The domain array managing communicator topology
        """
        self.mpi_domain_array = mpi_domain_array
        self.subdomain_comm = mpi_domain_array.getSubdomainComm()
        self.my_domain_index = mpi_domain_array.getDomainIndex()
        self.my_subdomain_index = mpi_domain_array.getSubdomainIndex()

    def _get_peer_rank(self, peer_domain_index):
        """
        Get the subdomain communicator rank for a peer domain.

        Since domains are arranged in a cartesian topology, the peer rank
        in the subdomain communicator corresponds to the peer domain index.
        """
        return peer_domain_index

    def send(self, data, dest_domain_index, tag=0):
        """
        Send a Data object to the corresponding rank in destination domain.

        This performs a rank-to-rank send where each rank sends its local portion
        of the Data to the same subdomain index in the destination domain.

        Parameters:
        -----------
        data : esys.escript.Data
            The Data object to send (will use local portion on this rank)
        dest_domain_index : int
            The index of the destination domain in the MPIDomainArray
        tag : int, optional
            MPI message tag (default: 0)

        Notes:
        ------
        - Data is automatically expanded and resolved before sending
        - Only the local portion on each rank is sent
        - Complex data is supported (sends real and imaginary parts separately)
        """
        if dest_domain_index == self.my_domain_index:
            raise ValueError("Cannot send to same domain (use local copy instead)")

        # Expand and resolve data before sending
        if not data.isExpanded():
            data.expand()
        if data.isLazy():
            data.resolve()

        # Get metadata
        shape = data.getShape()
        is_complex = data.isComplex()
        is_expanded = data.isExpanded()
        num_data_points = data.getNumberOfDataPoints()

        # Allocate numpy array for local data
        # Shape is (num_data_points,) + shape
        if len(shape) == 0:
            # Scalar data
            array_shape = (num_data_points,)
        else:
            # Vector/tensor data
            array_shape = (num_data_points,) + tuple(shape)

        if is_complex:
            local_data = np.empty(array_shape, dtype=np.complex128)
        else:
            local_data = np.empty(array_shape, dtype=np.float64)

        # Copy data point-by-point into numpy array
        for i in range(num_data_points):
            val = data.getTupleForDataPoint(i)
            # For scalar data, getTupleForDataPoint returns (value,) so extract it
            if len(shape) == 0:
                local_data[i] = val[0] if isinstance(val, tuple) else val
            else:
                local_data[i] = val

        # Flatten for sending
        local_data = local_data.flatten()

        # Get destination rank in subdomain communicator
        dest_rank = self._get_peer_rank(dest_domain_index)

        # Send metadata (shape, expanded flag, complex flag)
        metadata = {
            'shape': shape,
            'is_complex': is_complex,
            'is_expanded': is_expanded,
            'data_size': len(local_data)
        }
        self.subdomain_comm.send(metadata, dest=dest_rank, tag=tag)

        # Send data
        if is_complex:
            # Send real and imaginary parts separately
            self.subdomain_comm.Send(local_data.real, dest=dest_rank, tag=tag+1)
            self.subdomain_comm.Send(local_data.imag, dest=dest_rank, tag=tag+2)
        else:
            self.subdomain_comm.Send(local_data, dest=dest_rank, tag=tag+1)

    def receive(self, function_space, source_domain_index, tag=0):
        """
        Receive a Data object from the corresponding rank in source domain.

        This performs a rank-to-rank receive where each rank receives data
        from the same subdomain index in the source domain.

        Parameters:
        -----------
        function_space : esys.escript.FunctionSpace
            The FunctionSpace for the received Data (must be compatible with sender)
        source_domain_index : int
            The index of the source domain in the MPIDomainArray
        tag : int, optional
            MPI message tag (must match sender's tag, default: 0)

        Returns:
        --------
        data : esys.escript.Data
            The received Data object on the specified FunctionSpace

        Notes:
        ------
        - The FunctionSpace must have the same structure as the sender's
        - Each rank receives only its local portion
        - Returns a Data object with the same properties (expanded/constant, complex/real)
        """
        if source_domain_index == self.my_domain_index:
            raise ValueError("Cannot receive from same domain (use local copy instead)")

        # Get source rank in subdomain communicator
        source_rank = self._get_peer_rank(source_domain_index)

        # Receive metadata
        metadata = self.subdomain_comm.recv(source=source_rank, tag=tag)

        shape = metadata['shape']
        is_complex = metadata['is_complex']
        is_expanded = metadata['is_expanded']
        data_size = metadata['data_size']

        # Receive data
        if is_complex:
            real_part = np.empty(data_size, dtype=np.float64)
            imag_part = np.empty(data_size, dtype=np.float64)
            self.subdomain_comm.Recv(real_part, source=source_rank, tag=tag+1)
            self.subdomain_comm.Recv(imag_part, source=source_rank, tag=tag+2)
            local_data = real_part + 1j * imag_part
        else:
            local_data = np.empty(data_size, dtype=np.float64)
            self.subdomain_comm.Recv(local_data, source=source_rank, tag=tag+1)

        # Reshape received data
        if len(shape) == 0:
            # Scalar data
            array_shape = (data_size,)
        else:
            # Vector/tensor data
            array_shape = (data_size // np.prod(shape),) + tuple(shape)

        local_data = local_data.reshape(array_shape)

        # Create empty Data object and populate point-by-point
        received_data = Data(0.0, shape, function_space)
        # Expand to allow point-by-point setting
        received_data.expand()

        num_data_points = received_data.getNumberOfDataPoints()

        for i in range(num_data_points):
            val_tuple = tuple(local_data[i]) if len(shape) > 0 else local_data[i].item()
            received_data.setValueOfDataPoint(i, val_tuple)

        return received_data

    def exchange(self, send_data, recv_function_space, peer_domain_index, tag=0):
        """
        Bidirectional exchange of Data between two domains.

        This is a convenience method for simultaneous send and receive.
        Both domains must use the same tag value.

        Parameters:
        -----------
        send_data : esys.escript.Data
            Data to send
        recv_function_space : esys.escript.FunctionSpace
            FunctionSpace for received data
        peer_domain_index : int
            Index of the peer domain
        tag : int, optional
            Base tag for messages (default: 0)
            Note: Uses tag, tag+1, tag+2 internally

        Returns:
        --------
        received_data : esys.escript.Data
            The received Data

        Notes:
        ------
        - Both domains must call this method with the same tag
        - Operations are ordered by domain index to avoid deadlock
        """
        if peer_domain_index == self.my_domain_index:
            raise ValueError("Cannot exchange with same domain (use local copy instead)")

        # Order operations by domain index to avoid deadlock
        if self.my_domain_index < peer_domain_index:
            # Lower index: send first, then receive
            self.send(send_data, peer_domain_index, tag=tag)
            received_data = self.receive(recv_function_space, peer_domain_index, tag=tag)
        else:
            # Higher index: receive first, then send
            received_data = self.receive(recv_function_space, peer_domain_index, tag=tag)
            self.send(send_data, peer_domain_index, tag=tag)

        return received_data

    def broadcast_value(self, value=None, root_domain_index=0):
        """
        Broadcast a scalar or numpy array from root domain to all other domains.

        The value is replicated across all ranks in all domains, maintaining
        the escript invariant that scalars/arrays are identical across all subdomains.

        Parameters:
        -----------
        value : float, int, numpy.ndarray, or None
            The value to broadcast (only used on rank 0 of root domain)
        root_domain_index : int, optional
            The index of the root domain (default: 0)

        Returns:
        --------
        result : float, int, or numpy.ndarray
            The broadcasted value (same on all ranks of all domains)

        Example::

            # Broadcast a scalar
            if my_domain_idx == 0:
                max_val = coupler.broadcast_value(42.0, root_domain_index=0)
            else:
                max_val = coupler.broadcast_value(root_domain_index=0)

            # Broadcast an array
            if my_domain_idx == 0:
                coeffs = coupler.broadcast_value(np.array([1.0, 2.0, 3.0]), root_domain_index=0)
            else:
                coeffs = coupler.broadcast_value(root_domain_index=0)
        """
        domain_comm = self.mpi_domain_array.getDomainComm()

        # First, broadcast from root rank 0 to all corresponding rank 0s via subdomain comm
        if domain_comm.Get_rank() == 0:
            result = self.subdomain_comm.bcast(value, root=root_domain_index)
        else:
            result = None

        # Then broadcast within each domain to replicate across all subdomains
        result = domain_comm.bcast(result, root=0)

        return result

    def allreduce_value(self, value, op=MPI.SUM):
        """
        All-reduce a scalar or numpy array across all domains.

        Since the value is the same on all ranks within a domain (escript invariant),
        only rank 0 of each domain participates in the reduction, then the result
        is broadcast within each domain.

        Parameters:
        -----------
        value : float, int, or numpy.ndarray
            The value to reduce (should be same on all ranks within domain)
        op : MPI.Op, optional
            MPI reduction operation (default: MPI.SUM)
            Options: MPI.SUM, MPI.MAX, MPI.MIN, MPI.PROD

        Returns:
        --------
        result : float, int, or numpy.ndarray
            Reduced value (same on all ranks of all domains)

        Example::

            # Reduce scalars from all domains
            my_max_temp = Lsup(temperature)  # Same on all ranks within domain
            global_max_temp = coupler.allreduce_value(my_max_temp, op=MPI.MAX)

            # Reduce arrays from all domains
            my_histogram = compute_histogram(data)  # Same on all ranks within domain
            total_histogram = coupler.allreduce_value(my_histogram, op=MPI.SUM)
        """
        domain_comm = self.mpi_domain_array.getDomainComm()

        # Only rank 0 of each domain participates in subdomain reduction
        if domain_comm.Get_rank() == 0:
            if isinstance(value, np.ndarray):
                # Use Allreduce for numpy arrays
                result = np.empty_like(value)
                self.subdomain_comm.Allreduce(value, result, op=op)
            else:
                # Use allreduce for scalars
                result = self.subdomain_comm.allreduce(value, op=op)
        else:
            result = None

        # Broadcast result to all ranks within each domain
        result = domain_comm.bcast(result, root=0)

        return result

    def broadcast(self, data=None, function_space=None, root_domain_index=0):
        """
        Broadcast a Data object from root domain to all other domains.

        This performs a collective broadcast where each rank in the root domain
        broadcasts its local portion to the corresponding rank in all other domains.

        Parameters:
        -----------
        data : esys.escript.Data or None
            The Data object to broadcast (only used on root domain)
        function_space : esys.escript.FunctionSpace or None
            The FunctionSpace for received Data (only used on non-root domains)
        root_domain_index : int, optional
            The index of the root domain (default: 0)

        Returns:
        --------
        result : esys.escript.Data
            The broadcasted Data object (same on all domains)

        Example::

            # Domain 0 broadcasts initial conditions to all ensemble members
            if my_domain_idx == 0:
                initial_temp = Scalar(20.0 + compute_perturbation(), Solution(domain))
                temp = coupler.broadcast(data=initial_temp, root_domain_index=0)
            else:
                temp = coupler.broadcast(function_space=Solution(domain), root_domain_index=0)

        Notes:
        ------
        - This is a collective operation - all domains must call it
        - Root domain must provide 'data', non-root domains must provide 'function_space'
        - All Data/FunctionSpace must have compatible structure
        - Each rank independently broadcasts with corresponding ranks in other domains
        - Works with multi-rank domains via subdomain communicator
        """
        is_root = (self.my_domain_index == root_domain_index)

        if is_root:
            if data is None:
                raise ValueError("Root domain must provide 'data' parameter")

            # Expand and resolve data before broadcasting
            if not data.isExpanded():
                data.expand()
            if data.isLazy():
                data.resolve()

            # Get metadata
            shape = data.getShape()
            is_complex = data.isComplex()
            num_data_points = data.getNumberOfDataPoints()
            function_space = data.getFunctionSpace()

        else:
            if function_space is None:
                raise ValueError("Non-root domains must provide 'function_space' parameter")

            # These will be received via broadcast
            shape = None
            is_complex = None
            num_data_points = None

        # Broadcast metadata from root to all domains
        metadata = {
            'shape': shape if is_root else None,
            'is_complex': is_complex if is_root else None,
            'num_data_points': num_data_points if is_root else None
        }
        metadata = self.subdomain_comm.bcast(metadata, root=root_domain_index)

        shape = metadata['shape']
        is_complex = metadata['is_complex']
        num_data_points = metadata['num_data_points']

        # Check if complex (not yet supported)
        if is_complex:
            raise NotImplementedError("Broadcast not yet implemented for complex Data")

        # Allocate numpy array for data
        if len(shape) == 0:
            array_shape = (num_data_points,)
        else:
            array_shape = (num_data_points,) + tuple(shape)

        local_array = np.empty(array_shape, dtype=np.float64)

        if is_root:
            # Extract local values from Data
            for i in range(num_data_points):
                val = data.getTupleForDataPoint(i)
                if len(shape) == 0:
                    local_array[i] = val[0] if isinstance(val, tuple) else val
                else:
                    local_array[i] = val

        # Broadcast array from root to all domains via subdomain communicator
        self.subdomain_comm.Bcast(local_array, root=root_domain_index)

        # Reconstruct Data object from broadcasted values
        result = Data(0.0, shape, function_space)
        result.expand()

        for i in range(num_data_points):
            val_tuple = tuple(local_array[i]) if len(shape) > 0 else local_array[i].item()
            result.setValueOfDataPoint(i, val_tuple)

        return result

    def allreduce(self, data, op=MPI.SUM):
        """
        All-reduce Data point-by-point across all domains in the subdomain communicator.

        This performs a collective reduction where each data point is reduced independently
        across all domains. Each rank reduces its local portion of Data with the corresponding
        rank in all other domains. The result has the same FunctionSpace and structure as the input.

        Parameters:
        -----------
        data : esys.escript.Data
            The Data object to reduce (local portion on this rank)
        op : MPI.Op, optional
            MPI reduction operation (default: MPI.SUM)
            Options: MPI.SUM, MPI.MAX, MPI.MIN, MPI.PROD

        Returns:
        --------
        result : esys.escript.Data
            Reduced Data object on the same FunctionSpace (available on all domains)

        Example::

            # Sum temperature fields from all ensemble members
            coupler = DataCoupler(domain_array)
            my_temperature = Scalar(value_array, Solution(domain))
            total_temperature = coupler.allreduce(my_temperature, op=MPI.SUM)
            avg_temperature = total_temperature / num_domains

        Notes:
        ------
        - This is a collective operation - all domains must call it
        - All Data objects must have compatible shapes and function spaces
        - Reduction happens point-by-point across domains via subdomain communicator
        - Each rank independently reduces with corresponding ranks in other domains
        - Works with multi-rank domains (each rank reduces its local portion)
        """
        # Expand and resolve data before reduction
        if not data.isExpanded():
            data.expand()
        if data.isLazy():
            data.resolve()

        # Check if complex (not yet supported for allreduce)
        if data.isComplex():
            raise NotImplementedError("All-reduce not yet implemented for complex Data")

        # Get the FunctionSpace for reconstruction
        function_space = data.getFunctionSpace()

        # Each rank independently reduces its local data with corresponding
        # ranks in other domains via the subdomain communicator

        # Get number of data points
        num_data_points = data.getNumberOfDataPoints()
        shape = data.getShape()

        # Allocate numpy array for local data
        if len(shape) == 0:
            # Scalar data
            array_shape = (num_data_points,)
        else:
            # Vector/tensor data
            array_shape = (num_data_points,) + tuple(shape)

        local_array = np.empty(array_shape, dtype=np.float64)

        # Extract all local values point-by-point
        for i in range(num_data_points):
            val = data.getTupleForDataPoint(i)
            # For scalar data, getTupleForDataPoint returns (value,) so extract it
            if len(shape) == 0:
                local_array[i] = val[0] if isinstance(val, tuple) else val
            else:
                local_array[i] = val

        # Allreduce across subdomain communicator
        reduced_array = np.empty_like(local_array)
        self.subdomain_comm.Allreduce(local_array, reduced_array, op=op)

        # Reconstruct Data object from reduced values
        result = Data(0.0, shape, function_space)
        # Expand to allow point-by-point setting
        result.expand()

        for i in range(num_data_points):
            val_tuple = tuple(reduced_array[i]) if len(shape) > 0 else reduced_array[i].item()
            result.setValueOfDataPoint(i, val_tuple)

        return result



# -*- coding: utf-8 -*-

##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################

"""
Tensor operations for escript.

This module provides tensor construction, algebra, arithmetic, and product
operations for escript Data objects, numpy arrays, and symbolic expressions.
"""


__copyright__ = """Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__ = """Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""
__url__ = "https://github.com/LutzGross/esys-escript.github.io"

import numpy

from . import escriptcpp as escore
from .escriptcpp import C_GeneralTensorProduct
from .start import HAVE_SYMBOLS
from .util_math import sqrt, wherePositive, whereNegative, whereNonPositive, whereNonNegative

if HAVE_SYMBOLS:
    from . import symboliccore as sym
else:
    class sym:
        Symbol = type(memoryview(b'1'))

__all__ = [
    # Construction
    'kronecker', 'identity', 'zeros', 'identityTensor', 'identityTensor4', 'unitVector',
    # Algebra
    'length', 'trace', 'transpose', 'swap_axes', 'symmetric', 'nonsymmetric',
    'antisymmetric', 'hermitian', 'antihermitian', 'inverse', 'escript_inverse',
    'eigenvalues', 'eigenvalues_and_eigenvectors',
    # Arithmetic
    'mult', 'maximum', 'minimum', 'clip', 'cross',
    # Products
    'inner', 'outer', 'matrixmult', 'matrix_mult', 'tensormult', 'tensor_mult',
    'generalTensorProduct', 'escript_generalTensorProduct',
    'transposed_matrix_mult', 'transposed_tensor_mult',
    'generalTransposedTensorProduct', 'escript_generalTransposedTensorProduct',
    'matrix_transposed_mult', 'tensor_transposed_mult',
    'generalTensorTransposedProduct', 'escript_generalTensorTransposedProduct'
]


# =========================================================================
#   Tensor Construction
# =========================================================================

def kronecker(d=3):
    """
    Returns the kronecker delta-symbol.

    :param d: dimension or an object that has the ``getDim`` method defining the
              dimension
    :type d: ``int``, `escript.Domain` or `escript.FunctionSpace`
    :return: the object u of rank 2 with *u[i,j]=1* for *i=j* and *u[i,j]=0*
             otherwise
    :rtype: ``numpy.ndarray`` or `escript.Data` of rank 2
    """
    return identityTensor(d)


def identity(shape=()):
    """
    Returns the ``shape`` x ``shape`` identity tensor.

    :param shape: input shape for the identity tensor
    :type shape: ``tuple`` of ``int``
    :return: array whose shape is shape x shape where *u[i,k]=1* for *i=k* and
             *u[i,k]=0* otherwise for len(shape)=1. If len(shape)=2:
             *u[i,j,k,l]=1* for *i=k and j=l* and *u[i,j,k,l]=0* otherwise.
    :rtype: ``numpy.ndarray`` of rank 1, rank 2 or rank 4
    :raise ValueError: if len(shape)>2
    """
    if len(shape) > 0:
        out = numpy.zeros(shape + shape, numpy.float64)
        if len(shape) == 1:
            for i0 in range(shape[0]):
                out[i0, i0] = 1.
        elif len(shape) == 2:
            for i0 in range(shape[0]):
                for i1 in range(shape[1]):
                    out[i0, i1, i0, i1] = 1.
        else:
            raise ValueError("identity: length of shape is restricted to 2.")
    else:
        out = 1.
    return out


def zeros(shape=()):
    """
    Returns the ``shape`` zero tensor.

    :param shape: input shape for the zero tensor
    :type shape: ``tuple`` of ``int``
    :return: array of shape filled with zeros
    :rtype: ``numpy.ndarray``
    """
    if len(shape) > 0:
        out = numpy.zeros(shape, numpy.float64)
    else:
        out = 0.
    return out


def identityTensor(d=3):
    """
    Returns the ``d`` x ``d`` identity matrix.

    :param d: dimension or an object that has the ``getDim`` method defining the
              dimension
    :type d: ``int``, `escript.Domain` or `escript.FunctionSpace`
    :return: the object u of rank 2 with *u[i,j]=1* for *i=j* and *u[i,j]=0*
             otherwise
    :rtype: ``numpy.ndarray`` or `escript.Data` of rank 2
    """
    if isinstance(d, escore.FunctionSpace):
        return escore.Data(identity((d.getDim(),)), d)
    elif isinstance(d, escore.Domain):
        return identity((d.getDim(),))
    else:
        return identity((d,))


def identityTensor4(d=3):
    """
    Returns the ``d`` x ``d`` x ``d`` x ``d`` identity tensor.

    :param d: dimension or an object that has the ``getDim`` method defining the
              dimension
    :type d: ``int`` or any object with a ``getDim`` method
    :return: the object u of rank 4 with *u[i,j,k,l]=1* for *i=k and j=l* and
             *u[i,j,k,l]=0* otherwise
    :rtype: ``numpy.ndarray`` or `escript.Data` of rank 4
    """
    if isinstance(d, escore.FunctionSpace):
        return escore.Data(identity((d.getDim(), d.getDim())), d)
    elif isinstance(d, escore.Domain):
        return identity((d.getDim(), d.getDim()))
    else:
        return identity((d, d))


def unitVector(i=0, d=3):
    """
    Returns a unit vector u of dimension d whose non-zero element is at index i.

    :param i: index for non-zero element
    :type i: ``int``
    :param d: dimension or an object that has the ``getDim`` method defining the
              dimension
    :type d: ``int``, `escript.Domain` or `escript.FunctionSpace`
    :return: the object u of rank 1 with *u[j]=1* for *j=index* and *u[j]=0*
             otherwise
    :rtype: ``numpy.ndarray`` or `escript.Data` of rank 1
    """
    return kronecker(d)[i]


# =========================================================================
#   Tensor Algebra
# =========================================================================

def length(arg):
    """
    Returns the length (Euclidean norm) of argument ``arg`` at each data point.

    :param arg: argument
    :type arg: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray``
    :rtype: ``float``, `escript.Data`, `Symbol` depending on the type of ``arg``
    """
    a = abs(arg)
    return sqrt(inner(a, a))


def trace(arg, axis_offset=0):
    """
    Returns the trace of ``arg`` which is the sum of ``arg[k,k]`` over k.

    :param arg: argument
    :type arg: `escript.Data`, `Symbol`, ``numpy.ndarray``
    :param axis_offset: ``axis_offset`` to components to sum over. ``axis_offset``
                        must be non-negative and less than the rank of ``arg`` +1.
                        The dimensions of component ``axis_offset`` and
                        axis_offset+1 must be equal.
    :type axis_offset: ``int``
    :return: trace of arg. The rank of the returned object is rank of ``arg``
             minus 2.
    :rtype: `escript.Data`, `Symbol` or ``numpy.ndarray`` depending on the type
            of ``arg``
    """
    if isinstance(arg, numpy.ndarray):
        sh = arg.shape
        if len(sh) < 2:
            raise ValueError("rank of argument must be greater than 1")
        if axis_offset < 0 or axis_offset > len(sh) - 2:
            raise ValueError("axis_offset must be between 0 and %d" % (len(sh) - 2))
        s1 = 1
        for i in range(axis_offset):
            s1 *= sh[i]
        s2 = 1
        for i in range(axis_offset + 2, len(sh)):
            s2 *= sh[i]
        if not sh[axis_offset] == sh[axis_offset + 1]:
            raise ValueError("dimensions of component %d and %d must match." % (axis_offset, axis_offset + 1))
        arg_reshaped = numpy.reshape(arg, (s1, sh[axis_offset], sh[axis_offset], s2))
        out = numpy.zeros([s1, s2], numpy.double)
        for i1 in range(s1):
            for i2 in range(s2):
                for j in range(sh[axis_offset]):
                    out[i1, i2] += arg_reshaped[i1, j, j, i2]
        out.resize(sh[:axis_offset] + sh[axis_offset + 2:])
        return out
    elif isinstance(arg, escore.Data):
        if arg.getRank() < 2:
            raise ValueError("rank of argument must be greater than 1")
        if axis_offset < 0 or axis_offset > arg.getRank() - 2:
            raise ValueError("axis_offset must be between 0 and %d" % (arg.getRank() - 2))
        s = list(arg.getShape())
        if not s[axis_offset] == s[axis_offset + 1]:
            raise ValueError("dimensions of component %d and %d must match." % (axis_offset, axis_offset + 1))
        return arg._trace(axis_offset)
    elif isinstance(arg, sym.Symbol):
        if arg.getRank() < 2:
            raise ValueError("rank of argument must be greater than 1")
        if axis_offset < 0 or axis_offset > arg.getRank() - 2:
            raise ValueError("axis_offset must be between 0 and %d" % (arg.getRank() - 2))
        s = list(arg.getShape())
        if not s[axis_offset] == s[axis_offset + 1]:
            raise ValueError("dimensions of component %d and %d must match." % (axis_offset, axis_offset + 1))
        return arg.trace(axis_offset)
    elif isinstance(arg, complex):
        raise TypeError("illegal argument type complex.")
    elif isinstance(arg, float):
        raise TypeError("illegal argument type float.")
    elif isinstance(arg, int):
        raise TypeError("illegal argument type int.")
    else:
        raise TypeError("Unknown argument type.")


def transpose(arg, axis_offset=None):
    """
    Returns the transpose of ``arg`` by swapping the first ``axis_offset`` and
    the last ``rank-axis_offset`` components.

    :param arg: argument
    :type arg: `escript.Data`, `Symbol`, ``numpy.ndarray``, ``float``, ``int``
    :param axis_offset: the first ``axis_offset`` components are swapped with the
                        rest. ``axis_offset`` must be non-negative and less or
                        equal to the rank of ``arg``. If ``axis_offset`` is not
                        present ``int(r/2)`` where r is the rank of ``arg`` is
                        used.
    :type axis_offset: ``int``
    :return: transpose of ``arg``
    :rtype: `escript.Data`, `Symbol`, ``numpy.ndarray``, ``float``, ``int``
            depending on the type of ``arg``
    """
    if isinstance(arg, numpy.ndarray):
        if axis_offset is None:
            axis_offset = int(arg.ndim / 2)
        return numpy.transpose(arg, axes=list(range(axis_offset, arg.ndim)) + list(range(0, axis_offset)))
    elif isinstance(arg, escore.Data):
        r = arg.getRank()
        if axis_offset is None:
            axis_offset = int(r / 2)
        if axis_offset < 0 or axis_offset > r:
            raise ValueError("axis_offset must be between 0 and %s" % r)
        return arg._transpose(axis_offset)
    elif isinstance(arg, complex):
        if not (axis_offset == 0 or axis_offset is None):
            raise ValueError("axis_offset must be 0 for complex argument")
        return arg
    elif isinstance(arg, float):
        if not (axis_offset == 0 or axis_offset is None):
            raise ValueError("axis_offset must be 0 for float argument")
        return arg
    elif isinstance(arg, int):
        if not (axis_offset == 0 or axis_offset is None):
            raise ValueError("axis_offset must be 0 for int argument")
        return float(arg)
    elif isinstance(arg, sym.Symbol):
        r = arg.getRank()
        if axis_offset is None:
            axis_offset = int(r / 2)
        if axis_offset < 0 or axis_offset > r:
            raise ValueError("axis_offset must be between 0 and %s" % r)
        return arg.transpose(axis_offset)
    else:
        raise TypeError("Unknown argument type.")


def swap_axes(arg, axis0=0, axis1=1):
    """
    Returns the swap of ``arg`` by swapping the components ``axis0`` and ``axis1``.

    :param arg: argument
    :type arg: `escript.Data`, `Symbol`, ``numpy.ndarray``
    :param axis0: first axis. ``axis0`` must be non-negative and less than the
                  rank of ``arg``.
    :type axis0: ``int``
    :param axis1: second axis. ``axis1`` must be non-negative and less than the
                  rank of ``arg``.
    :type axis1: ``int``
    :return: ``arg`` with swapped components
    :rtype: `escript.Data`, `Symbol` or ``numpy.ndarray`` depending on the type
            of ``arg``
    """
    if axis0 > axis1:
        axis0, axis1 = axis1, axis0
    if isinstance(arg, numpy.ndarray):
        return numpy.swapaxes(arg, axis0, axis1)
    elif isinstance(arg, escore.Data):
        return arg._swap_axes(axis0, axis1)
    elif isinstance(arg, sym.Symbol):
        return arg.swap_axes(axis0, axis1)
    elif isinstance(arg, complex):
        raise TypeError("complex argument is not supported.")
    elif isinstance(arg, float):
        raise TypeError("float argument is not supported.")
    elif isinstance(arg, int):
        raise TypeError("int argument is not supported.")
    else:
        raise TypeError("Unknown argument type.")


def symmetric(arg):
    """
    Returns the symmetric part of the square matrix ``arg``. That is,
    *(arg+transpose(arg))/2*.

    :param arg: input matrix. Must have rank 2 or 4 and be square.
    :type arg: ``numpy.ndarray``, `escript.Data`, `Symbol`
    :return: symmetric part of ``arg``
    :rtype: ``numpy.ndarray``, `escript.Data`, `Symbol` depending on the input
    """
    if isinstance(arg, numpy.ndarray):
        if arg.ndim == 2:
            if not (arg.shape[0] == arg.shape[1]):
                raise ValueError("argument must be square.")
        elif arg.ndim == 4:
            if not (arg.shape[0] == arg.shape[2] and arg.shape[1] == arg.shape[3]):
                raise ValueError("argument must be square.")
        else:
            raise ValueError("rank 2 or 4 is required.")
        return (arg + transpose(arg)) / 2
    elif isinstance(arg, escore.Data):
        if arg.getRank() == 2:
            if not (arg.getShape()[0] == arg.getShape()[1]):
                raise ValueError("argument must be square.")
            return arg._symmetric()
        elif arg.getRank() == 4:
            if not (arg.getShape()[0] == arg.getShape()[2] and arg.getShape()[1] == arg.getShape()[3]):
                raise ValueError("argument must be square.")
            return arg._symmetric()
        else:
            raise ValueError("rank 2 or 4 is required.")
    elif isinstance(arg, sym.Symbol):
        if arg.getRank() == 2:
            if arg.getShape()[0] != arg.getShape()[1]:
                raise ValueError("symmetric: argument must be square.")
        elif arg.getRank() == 4:
            if arg.getShape()[0] != arg.getShape()[2] or arg.getShape()[1] != arg.getShape()[3]:
                raise ValueError("symmetric: argument must be square.")
        else:
            raise ValueError("symmetric: rank 2 or 4 is required.")
        return (arg + transpose(arg)) / 2
    elif isinstance(arg, complex):
        return arg
    elif isinstance(arg, float):
        return arg
    elif isinstance(arg, int):
        return float(arg)
    else:
        raise TypeError("symmetric: Unknown argument type.")


def nonsymmetric(arg):
    """
    Deprecated alias for antisymmetric.
    """
    return antisymmetric(arg)


def antisymmetric(arg):
    """
    Returns the anti-symmetric part of the square matrix ``arg``. That is,
    *(arg-transpose(arg))/2*.

    :param arg: input matrix. Must have rank 2 or 4 and be square.
    :type arg: ``numpy.ndarray``, `escript.Data`, `Symbol`
    :return: anti-symmetric part of ``arg``
    :rtype: ``numpy.ndarray``, `escript.Data`, `Symbol` depending on the input
    """
    if isinstance(arg, numpy.ndarray):
        if arg.ndim == 2:
            if not (arg.shape[0] == arg.shape[1]):
                raise ValueError("antisymmetric: argument must be square.")
        elif arg.ndim == 4:
            if not (arg.shape[0] == arg.shape[2] and arg.shape[1] == arg.shape[3]):
                raise ValueError("antisymmetric: argument must be square.")
        else:
            raise ValueError("antisymmetric: rank 2 or 4 is required.")
        return (arg - transpose(arg)) / 2
    elif isinstance(arg, escore.Data):
        if arg.getRank() == 2:
            if not (arg.getShape()[0] == arg.getShape()[1]):
                raise ValueError("argument must be square.")
            return arg._antisymmetric()
        elif arg.getRank() == 4:
            if not (arg.getShape()[0] == arg.getShape()[2] and arg.getShape()[1] == arg.getShape()[3]):
                raise ValueError("argument must be square.")
            return arg._antisymmetric()
        else:
            raise ValueError("rank 2 or 4 is required.")
    elif isinstance(arg, sym.Symbol):
        if arg.getRank() == 2:
            if arg.getShape()[0] != arg.getShape()[1]:
                raise ValueError("antisymmetric: argument must be square.")
        elif arg.getRank() == 4:
            if arg.getShape()[0] != arg.getShape()[2] or arg.getShape()[1] != arg.getShape()[3]:
                raise ValueError("antisymmetric: argument must be square.")
        else:
            raise ValueError("antisymmetric: rank 2 or 4 is required.")
        return (arg - transpose(arg)) / 2
    elif isinstance(arg, complex):
        return complex(0)
    elif isinstance(arg, float):
        return float(0)
    elif isinstance(arg, int):
        return float(0)
    else:
        raise TypeError("antisymmetric: Unknown argument type.")


def hermitian(arg):
    """
    Returns the hermitian part of the square matrix ``arg``. That is,
    *(arg+adjoint(arg))/2*.

    :param arg: input matrix. Must have rank 2 or 4 and be square.
    :type arg: ``numpy.ndarray``, `escript.Data`, `Symbol`
    :return: hermitian part of ``arg``
    :rtype: ``numpy.ndarray``, `escript.Data`, `Symbol` depending on the input
    """
    if isinstance(arg, numpy.ndarray):
        if arg.ndim == 2:
            if not (arg.shape[0] == arg.shape[1]):
                raise ValueError("argument must be square.")
        elif arg.ndim == 4:
            if not (arg.shape[0] == arg.shape[2] and arg.shape[1] == arg.shape[3]):
                raise ValueError("argument must be square.")
        else:
            raise ValueError("rank 2 or 4 is required.")
        return (arg + transpose(arg).conj()) / 2
    elif isinstance(arg, escore.Data):
        if arg.getRank() == 2:
            if not (arg.getShape()[0] == arg.getShape()[1]):
                raise ValueError("argument must be square.")
            return arg._hermitian()
        elif arg.getRank() == 4:
            if not (arg.getShape()[0] == arg.getShape()[2] and arg.getShape()[1] == arg.getShape()[3]):
                raise ValueError("argument must be square.")
            return arg._hermitian()
        else:
            raise ValueError("rank 2 or 4 is required.")
    elif isinstance(arg, sym.Symbol):
        if arg.getRank() == 2:
            if arg.getShape()[0] != arg.getShape()[1]:
                raise ValueError("hermitian: argument must be square.")
        elif arg.getRank() == 4:
            if arg.getShape()[0] != arg.getShape()[2] or arg.getShape()[1] != arg.getShape()[3]:
                raise ValueError("hermitian: argument must be square.")
        else:
            raise ValueError("hermitian: rank 2 or 4 is required.")
        # Note: adjoint not available, using transpose as approximation
        return (arg + transpose(arg)) / 2
    elif isinstance(arg, complex):
        return complex(arg.real)
    elif isinstance(arg, float):
        return arg
    elif isinstance(arg, int):
        return float(arg)
    else:
        raise TypeError("hermitian: Unknown argument type.")


def antihermitian(arg):
    """
    Returns the anti-hermitian part of the square matrix ``arg``. That is,
    *(arg-adjoint(arg))/2*.

    :param arg: input matrix. Must have rank 2 or 4 and be square.
    :type arg: ``numpy.ndarray``, `escript.Data`, `Symbol`
    :return: anti-hermitian part of ``arg``
    :rtype: ``numpy.ndarray``, `escript.Data`, `Symbol` depending on the input
    """
    if isinstance(arg, numpy.ndarray):
        if arg.ndim == 2:
            if not (arg.shape[0] == arg.shape[1]):
                raise ValueError("antihermitian: argument must be square.")
        elif arg.ndim == 4:
            if not (arg.shape[0] == arg.shape[2] and arg.shape[1] == arg.shape[3]):
                raise ValueError("antihermitian: argument must be square.")
        else:
            raise ValueError("antihermitian: rank 2 or 4 is required.")
        return (arg - transpose(arg).conj()) / 2
    elif isinstance(arg, escore.Data):
        if arg.getRank() == 2:
            if not (arg.getShape()[0] == arg.getShape()[1]):
                raise ValueError("argument must be square.")
            return arg._antihermitian()
        elif arg.getRank() == 4:
            if not (arg.getShape()[0] == arg.getShape()[2] and arg.getShape()[1] == arg.getShape()[3]):
                raise ValueError("argument must be square.")
            return arg._antihermitian()
        else:
            raise ValueError("rank 2 or 4 is required.")
    elif isinstance(arg, sym.Symbol):
        if arg.getRank() == 2:
            if arg.getShape()[0] != arg.getShape()[1]:
                raise ValueError("antihermitian: argument must be square.")
        elif arg.getRank() == 4:
            if arg.getShape()[0] != arg.getShape()[2] or arg.getShape()[1] != arg.getShape()[3]:
                raise ValueError("antihermitian: argument must be square.")
        else:
            raise ValueError("antihermitian: rank 2 or 4 is required.")
        return (arg - hermitian(arg)) / 2
    elif isinstance(arg, complex):
        return complex(arg.imag * 1j)
    elif isinstance(arg, float):
        return float(0)
    elif isinstance(arg, int):
        return float(0)
    else:
        raise TypeError("antihermitian: Unknown argument type.")


def inverse(arg):
    """
    Returns the inverse of the square matrix ``arg``.

    :param arg: square matrix. Must have rank 2 and the first and second
                dimension must be equal.
    :type arg: ``numpy.ndarray``, `escript.Data`, `Symbol`
    :return: inverse of the argument. ``matrix_mult(inverse(arg),arg)`` will be
             almost equal to ``kronecker(arg.getShape()[0])``
    :rtype: ``numpy.ndarray``, `escript.Data`, `Symbol` depending on the input
    :note: for `escript.Data` objects the dimension is restricted to 3.
    """
    import numpy.linalg
    if isinstance(arg, numpy.ndarray):
        return numpy.linalg.tensorinv(arg, ind=1)
    elif isinstance(arg, escore.Data):
        return escript_inverse(arg)
    elif isinstance(arg, complex):
        return 1. / arg
    elif isinstance(arg, float):
        return 1. / arg
    elif isinstance(arg, int):
        return 1. / float(arg)
    elif isinstance(arg, sym.Symbol):
        return arg.inverse()
    else:
        raise TypeError("inverse: Unknown argument type.")


def escript_inverse(arg):
    """
    Returns the inverse of an escript Data matrix.

    :param arg: square matrix Data object (dimension restricted to 3)
    :type arg: `escript.Data`
    :return: inverse of the matrix
    :rtype: `escript.Data`
    """
    return arg._inverse()


def eigenvalues(arg):
    """
    Returns the eigenvalues of the square matrix ``arg``.

    :param arg: square matrix. Must have rank 2 and the first and second
                dimension must be equal. It must also be symmetric, ie.
                ``transpose(arg)==arg`` (this is not checked).
    :type arg: ``numpy.ndarray``, `escript.Data`, `Symbol`
    :return: the eigenvalues in increasing order
    :rtype: ``numpy.ndarray``, `escript.Data`, `Symbol` depending on the input
    :note: for `escript.Data` and `Symbol` objects the dimension is
           restricted to 3.
    """
    if isinstance(arg, numpy.ndarray):
        out = numpy.linalg.eigvals((arg + numpy.transpose(arg)) / 2.)
        out.sort()
        return out
    elif isinstance(arg, escore.Data):
        return arg._eigenvalues()
    elif isinstance(arg, complex):
        return arg
    elif isinstance(arg, float):
        return arg
    elif isinstance(arg, int):
        return float(arg)
    elif isinstance(arg, sym.Symbol):
        return sym.symfn.eigenvalues(arg)
    else:
        raise TypeError("eigenvalues: Unknown argument type.")


def eigenvalues_and_eigenvectors(arg):
    """
    Returns the eigenvalues and eigenvectors of the square matrix ``arg``.

    :param arg: square matrix. Must have rank 2 and the first and second
                dimension must be equal. It must also be symmetric, ie.
                ``transpose(arg)==arg`` (this is not checked).
    :type arg: `escript.Data`
    :return: the eigenvalues and eigenvectors. The eigenvalues are ordered by
             increasing value. The eigenvectors are orthogonal and normalized.
             If V are the eigenvectors then V[:,i] is the eigenvector
             corresponding to the i-th eigenvalue.
    :rtype: `tuple` of `escript.Data`
    :note: The dimension is restricted to 3.
    """
    if isinstance(arg, numpy.ndarray):
        raise TypeError("eigenvalues_and_eigenvectors does not support numpy.ndarray arguments")
    elif isinstance(arg, escore.Data):
        return arg._eigenvalues_and_eigenvectors()
    elif isinstance(arg, complex):
        return (numpy.array([[arg]], numpy.cdouble_), numpy.ones((1, 1), numpy.cdouble_))
    elif isinstance(arg, float):
        return (numpy.array([[arg]], numpy.float_), numpy.ones((1, 1), numpy.float_))
    elif isinstance(arg, int):
        return (numpy.array([[arg]], numpy.float_), numpy.ones((1, 1), numpy.float_))
    elif isinstance(arg, sym.Symbol):
        return sym.symfn.eigenvalues_and_eigenvectors(arg)
    else:
        raise TypeError("eigenvalues: Unknown argument type.")


# =========================================================================
#   Tensor Arithmetic
# =========================================================================

def mult(arg0, arg1):
    """
    Product of ``arg0`` and ``arg1``.

    :param arg0: first term
    :type arg0: `Symbol`, ``float``, ``int``, `escript.Data` or ``numpy.ndarray``
    :param arg1: second term
    :type arg1: `Symbol`, ``float``, ``int``, `escript.Data` or ``numpy.ndarray``
    :return: the product of ``arg0`` and ``arg1``
    :rtype: `Symbol`, ``float``, ``int``, `escript.Data` or ``numpy.ndarray``
    :note: The shape of both arguments is matched according to the rules
           used in `matchShape`.
    """
    # Import locally to avoid circular import
    from . import util
    args = util.matchShape(arg0, arg1)
    if util.testForZero(args[0]) or util.testForZero(args[1]):
        return numpy.zeros(util.getShape(args[0]), numpy.double)
    else:
        if isinstance(args[0], numpy.ndarray):
            return args[1] * args[0]
        else:
            return args[0] * args[1]


def maximum(*args):
    """
    The maximum over arguments ``args``.

    :param args: arguments
    :type args: ``numpy.ndarray``, `escript.Data`, `Symbol`, ``int`` or ``float``
    :return: an object which in each entry gives the maximum of the
             corresponding values in ``args``
    :rtype: ``numpy.ndarray``, `escript.Data`, `Symbol`, ``int`` or ``float``
            depending on the input
    """
    if max([isinstance(v, sym.Symbol) for v in args]):
        return sym.symfn.maximum(*args)
    out = None
    for a in args:
        if out is None:
            out = a * 1.
        else:
            if isinstance(out, escore.Data) and isinstance(a, escore.Data):
                if out.getRank() == 0 and a.getRank() > 0:
                    res = a.copy()
                    res.copyWithMask(out, wherePositive(out - a))
                    out = res
                else:
                    out.copyWithMask(a, wherePositive(a - out))
            else:
                if isinstance(a, numpy.ndarray):
                    diff = -out + a
                else:
                    diff = a - out
                temp = mult(whereNonPositive(diff), out) + mult(wherePositive(diff), a)
                if isinstance(out, numpy.ndarray) and isinstance(a, numpy.ndarray):
                    temp = numpy.array(temp)
                out = temp
    return out


def minimum(*args):
    """
    The minimum over arguments ``args``.

    :param args: arguments
    :type args: ``numpy.ndarray``, `escript.Data`, `Symbol`, ``int`` or ``float``
    :return: an object which gives in each entry the minimum of the
             corresponding values in ``args``
    :rtype: ``numpy.ndarray``, `escript.Data`, `Symbol`, ``int`` or ``float``
            depending on the input
    """
    if max([isinstance(v, sym.Symbol) for v in args]):
        return sym.symfn.minimum(*args)
    out = None
    for a in args:
        if out is None:
            if isinstance(a, numpy.ndarray):
                out = a.copy()
            else:
                out = a * 1.
        else:
            if isinstance(out, escore.Data) and isinstance(a, escore.Data):
                if out.getRank() == 0 and a.getRank() > 0:
                    res = a.copy()
                    res.copyWithMask(out, whereNegative(out - a))
                    out = res
                else:
                    out.copyWithMask(a, whereNegative(a - out))
            else:
                if isinstance(a, numpy.ndarray):
                    diff = -out + a
                else:
                    diff = a - out
                temp = mult(whereNonNegative(diff), out) + mult(whereNegative(diff), a)
                if isinstance(out, numpy.ndarray) and isinstance(a, numpy.ndarray):
                    temp = numpy.array(temp)
                out = temp
    return out


def clip(arg, minval=None, maxval=None):
    """
    Cuts the values of ``arg`` between ``minval`` and ``maxval``.

    :param arg: argument
    :type arg: ``numpy.ndarray``, `escript.Data`, `Symbol`, ``int`` or ``float``
    :param minval: lower range. If None no lower range is applied
    :type minval: ``float`` or ``None``
    :param maxval: upper range. If None no upper range is applied
    :type maxval: ``float`` or ``None``
    :return: an object that contains all values from ``arg`` between ``minval``
             and ``maxval``
    :rtype: ``numpy.ndarray``, `escript.Data`, `Symbol`, ``int`` or ``float``
            depending on the input
    :raise ValueError: if ``minval>maxval``
    """
    if isinstance(arg, sym.Symbol):
        clip_item = lambda item: sym.symfn.clip(item, minval, maxval)
        return arg.applyfunc(clip_item)
    if minval is not None and maxval is not None:
        if minval > maxval:
            raise ValueError("minval = %s must be less than maxval %s" % (minval, maxval))
    if minval is None:
        tmp = arg
    else:
        tmp = maximum(minval, arg)
    if maxval is None:
        return tmp
    else:
        return minimum(tmp, maxval)


def cross(arg0, arg1):
    """
    Cross product of the two arguments ``arg0`` and ``arg1`` which need to be shape (3,).

    :param arg0: first argument
    :type arg0: ``numpy.ndarray``, `escript.Data`, `Symbol`
    :param arg1: second argument
    :type arg1: ``numpy.ndarray``, `escript.Data`, `Symbol`
    :return: the cross product of ``arg0`` and ``arg1`` at each data point
    :rtype: ``numpy.ndarray``, `escript.Data`, `Symbol` depending on ``arg0``
    :raise ValueError: if the shapes of the arguments are not (3,)
    """
    from . import util
    if isinstance(arg0, numpy.ndarray) and isinstance(arg1, numpy.ndarray):
        out = numpy.cross(arg0, arg1)
    else:
        sh0 = util.getShape(arg0)
        sh1 = util.getShape(arg1)
        if not sh0 == (3,):
            raise ValueError("cross: arg0 needs to be of shape (3,)")
        if not sh1 == (3,):
            raise ValueError("cross: arg1 needs to be of shape (3,)")

        if isinstance(arg0, sym.Symbol):
            if isinstance(arg1, sym.Symbol):
                out = sym.Symbol(arg0.name + "x" + arg1.name, (3,))
            else:
                out = sym.Symbol(arg0.name + "x" + str(type(arg1)), (3,))
        elif isinstance(arg0, escore.Data):
            if isinstance(arg1, sym.Symbol):
                out = sym.Symbol(str(type(arg0)) + "x" + arg1.name, (3,))
            else:
                out = escore.Data(0., (3,), arg0.getFunctionSpace())
        elif isinstance(arg1, escore.Data):
            if isinstance(arg0, sym.Symbol):
                out = sym.Symbol(str(type(arg0)) + "x" + arg1.name, (3,))
            else:
                out = escore.Data(0., (3,), arg1.getFunctionSpace())
        else:
            raise TypeError("cross: argument type not supported")

        out[0] = arg0[1] * arg1[2] - arg0[2] * arg1[1]
        out[1] = arg0[2] * arg1[0] - arg0[0] * arg1[2]
        out[2] = arg0[0] * arg1[1] - arg0[1] * arg1[0]
    return out


# =========================================================================
#   Tensor Products
# =========================================================================

def inner(arg0, arg1):
    """
    Inner product of the two arguments. The inner product is defined as:

    `out=Sigma_s arg0[s]*arg1[s]`

    where s runs through ``arg0.Shape``.

    ``arg0`` and ``arg1`` must have the same shape.

    :param arg0: first argument
    :type arg0: ``numpy.ndarray``, `escript.Data`, `Symbol`, ``float``, ``int``
    :param arg1: second argument
    :type arg1: ``numpy.ndarray``, `escript.Data`, `Symbol`, ``float``, ``int``
    :return: the inner product of ``arg0`` and ``arg1`` at each data point
    :rtype: ``numpy.ndarray``, `escript.Data`, `Symbol`, ``float``
            depending on the input
    :raise ValueError: if the shapes of the arguments are not identical
    """
    from . import util
    sh0 = util.getShape(arg0)
    sh1 = util.getShape(arg1)
    if not sh0 == sh1:
        raise ValueError("inner: shape of arguments does not match")
    return generalTensorProduct(arg0, arg1, axis_offset=len(sh0))


def outer(arg0, arg1):
    """
    The outer product of the two arguments. The outer product is defined as:

    ``out[t,s]=arg0[t]*arg1[s]``

    where
        - s runs through ``arg0.Shape``
        - t runs through ``arg1.Shape``

    :param arg0: first argument
    :type arg0: ``numpy.ndarray``, `escript.Data`, `Symbol`, ``float``, ``int``
    :param arg1: second argument
    :type arg1: ``numpy.ndarray``, `escript.Data`, `Symbol`, ``float``, ``int``
    :return: the outer product of ``arg0`` and ``arg1`` at each data point
    :rtype: ``numpy.ndarray``, `escript.Data`, `Symbol` depending on the input
    """
    return generalTensorProduct(arg0, arg1, axis_offset=0)


def matrixmult(arg0, arg1):
    """
    See `matrix_mult`.
    """
    return matrix_mult(arg0, arg1)


def matrix_mult(arg0, arg1):
    """
    matrix-matrix or matrix-vector product of the two arguments.

    `out[s0]=Sigma_{r0} arg0[s0,r0]*arg1[r0]`

    or

    `out[s0,s1]=Sigma_{r0} arg0[s0,r0]*arg1[r0,s1]`

    The second dimension of ``arg0`` and the first dimension of ``arg1`` must
    match.

    :param arg0: first argument of rank 2
    :type arg0: ``numpy.ndarray``, `escript.Data`, `Symbol`
    :param arg1: second argument of at least rank 1
    :type arg1: ``numpy.ndarray``, `escript.Data`, `Symbol`
    :return: the matrix-matrix or matrix-vector product of ``arg0`` and ``arg1``
             at each data point
    :rtype: ``numpy.ndarray``, `escript.Data`, `Symbol` depending on the input
    :raise ValueError: if the shapes of the arguments are not appropriate
    """
    from . import util
    sh0 = util.getShape(arg0)
    sh1 = util.getShape(arg1)
    if not len(sh0) == 2:
        raise ValueError("first argument must have rank 2")
    if not len(sh1) == 2 and not len(sh1) == 1:
        raise ValueError("second argument must have rank 1 or 2")
    return generalTensorProduct(arg0, arg1, axis_offset=1)


def tensormult(arg0, arg1):
    """
    See `tensor_mult`.
    """
    return tensor_mult(arg0, arg1)


def tensor_mult(arg0, arg1):
    """
    The tensor product of the two arguments.

    For ``arg0`` of rank 2 this is

    `out[s0]=Sigma_{r0} arg0[s0,r0]*arg1[r0]`

    or

    `out[s0,s1]=Sigma_{r0} arg0[s0,r0]*arg1[r0,s1]`

    and for ``arg0`` of rank 4 this is

    `out[s0,s1,s2,s3]=Sigma_{r0,r1} arg0[s0,s1,r0,r1]*arg1[r0,r1,s2,s3]`

    or

    `out[s0,s1,s2]=Sigma_{r0,r1} arg0[s0,s1,r0,r1]*arg1[r0,r1,s2]`

    or

    `out[s0,s1]=Sigma_{r0,r1} arg0[s0,s1,r0,r1]*arg1[r0,r1]`

    In the first case the second dimension of ``arg0`` and the last dimension of
    ``arg1`` must match and in the second case the two last dimensions of ``arg0``
    must match the two first dimensions of ``arg1``.

    :param arg0: first argument of rank 2 or 4
    :type arg0: ``numpy.ndarray``, `escript.Data`, `Symbol`
    :param arg1: second argument of shape greater than 1 or 2 depending on the
                 rank of ``arg0``
    :type arg1: ``numpy.ndarray``, `escript.Data`, `Symbol`
    :return: the tensor product of ``arg0`` and ``arg1`` at each data point
    :rtype: ``numpy.ndarray``, `escript.Data`, `Symbol` depending on the input
    """
    from . import util
    sh0 = util.getShape(arg0)
    sh1 = util.getShape(arg1)
    if len(sh0) == 2 and (len(sh1) == 2 or len(sh1) == 1):
        return generalTensorProduct(arg0, arg1, axis_offset=1)
    elif len(sh0) == 4 and (len(sh1) == 2 or len(sh1) == 3 or len(sh1) == 4):
        return generalTensorProduct(arg0, arg1, axis_offset=2)
    else:
        raise ValueError("tensor_mult: first argument must have rank 2 or 4 and second rank must be in (1,2) or (2,3,4) respectively.")


def generalTensorProduct(arg0, arg1, axis_offset=0):
    """
    Generalized tensor product.

    `out[s,t]=Sigma_r arg0[s,r]*arg1[r,t]`

    where
        - s runs through ``arg0.Shape[:arg0.ndim-axis_offset]``
        - r runs through ``arg1.Shape[:axis_offset]``
        - t runs through ``arg1.Shape[axis_offset:]``

    :param arg0: first argument
    :type arg0: ``numpy.ndarray``, `escript.Data`, `Symbol`, ``float``, ``int``
    :param arg1: second argument
    :type arg1: ``numpy.ndarray``, `escript.Data`, `Symbol`, ``float``, ``int``
    :return: the general tensor product of ``arg0`` and ``arg1`` at each data
             point
    :rtype: ``numpy.ndarray``, `escript.Data`, `Symbol` depending on the input
    """
    from . import util
    if (isinstance(arg0, float) or isinstance(arg0, complex)) and (isinstance(arg1, float) or isinstance(arg1, complex)):
        return arg1 * arg0
    arg0, arg1 = util.matchType(arg0, arg1)
    if isinstance(arg0, sym.Symbol):
        sh0 = arg0.getShape()
        sh1 = util.getShape(arg1)
        if not sh0[arg0.getRank() - axis_offset:] == sh1[:axis_offset]:
            raise ValueError("dimensions of last %s components in left argument don't match the first %s components in the right argument." % (axis_offset, axis_offset))
        if isinstance(arg1, float):
            return arg0 * arg1
        elif isinstance(arg1, numpy.ndarray) or isinstance(arg1, sym.Symbol):
            return arg0.tensorProduct(arg1, axis_offset)
        elif isinstance(arg1, escore.Data):
            raise TypeError("tensor product of Symbol and Data not supported yet")
    elif isinstance(arg0, numpy.ndarray):
        if not arg0.shape[arg0.ndim - axis_offset:] == arg1.shape[:axis_offset]:
            raise ValueError("dimensions of last %s components in left argument don't match the first %s components in the right argument." % (axis_offset, axis_offset))
        arg0_c = arg0.copy()
        arg1_c = arg1.copy()
        sh0, sh1 = arg0.shape, arg1.shape
        d0, d1, d01 = 1, 1, 1
        for i in sh0[:arg0.ndim - axis_offset]:
            d0 *= i
        for i in sh1[axis_offset:]:
            d1 *= i
        for i in sh1[:axis_offset]:
            d01 *= i
        arg0_c.resize((d0, d01))
        arg1_c.resize((d01, d1))
        if arg0_c.dtype.kind == 'c':
            restype = arg0_c.dtype
        else:
            restype = arg1_c.dtype
        out = numpy.zeros((d0, d1), restype)
        for i0 in range(d0):
            for i1 in range(d1):
                out[i0, i1] = numpy.sum(arg0_c[i0, :] * arg1_c[:, i1])
        out.resize(sh0[:arg0.ndim - axis_offset] + sh1[axis_offset:])
        return out
    elif isinstance(arg0, escore.Data):
        if isinstance(arg1, sym.Symbol):
            raise TypeError("tensor product of Data and Symbol not supported yet")
        return escript_generalTensorProduct(arg0, arg1, axis_offset)
    raise TypeError("generalTensorProduct: Unsupported argument type")


def escript_generalTensorProduct(arg0, arg1, axis_offset, transpose=0):
    """
    Generalized tensor product for escript Data objects.

    :param arg0: first Data object
    :type arg0: `escript.Data`
    :param arg1: second Data object (not necessarily on the same function space)
    :type arg1: `escript.Data`
    :param axis_offset: axis offset for the tensor product
    :type axis_offset: ``int``
    :param transpose: transpose mode (0=none, 1=transpose arg0, 2=transpose arg1)
    :type transpose: ``int``
    :return: tensor product result
    :rtype: `escript.Data`
    """
    return C_GeneralTensorProduct(arg0, arg1, axis_offset, transpose)


def transposed_matrix_mult(arg0, arg1):
    """
    transposed(matrix)-matrix or transposed(matrix)-vector product of the two
    arguments.

    `out[s0]=Sigma_{r0} arg0[r0,s0]*arg1[r0]`

    or

    `out[s0,s1]=Sigma_{r0} arg0[r0,s0]*arg1[r0,s1]`

    The function call ``transposed_matrix_mult(arg0,arg1)`` is equivalent to
    ``matrix_mult(transpose(arg0),arg1)``.

    The first dimension of ``arg0`` and ``arg1`` must match.

    :param arg0: first argument of rank 2
    :type arg0: ``numpy.ndarray``, `escript.Data`, `Symbol`
    :param arg1: second argument of at least rank 1
    :type arg1: ``numpy.ndarray``, `escript.Data`, `Symbol`
    :return: the product of the transpose of ``arg0`` and ``arg1`` at each data
             point
    :rtype: ``numpy.ndarray``, `escript.Data`, `Symbol` depending on the input
    :raise ValueError: if the shapes of the arguments are not appropriate
    """
    from . import util
    sh0 = util.getShape(arg0)
    sh1 = util.getShape(arg1)
    if not len(sh0) == 2:
        raise ValueError("first argument must have rank 2")
    if not len(sh1) == 2 and not len(sh1) == 1:
        raise ValueError("second argument must have rank 1 or 2")
    return generalTransposedTensorProduct(arg0, arg1, axis_offset=1)


def transposed_tensor_mult(arg0, arg1):
    """
    The tensor product of the transpose of the first and the second argument.

    For ``arg0`` of rank 2 this is

    `out[s0]=Sigma_{r0} arg0[r0,s0]*arg1[r0]`

    or

    `out[s0,s1]=Sigma_{r0} arg0[r0,s0]*arg1[r0,s1]`

    and for ``arg0`` of rank 4 this is

    `out[s0,s1,s2,s3]=Sigma_{r0,r1} arg0[r0,r1,s0,s1]*arg1[r0,r1,s2,s3]`

    or

    `out[s0,s1,s2]=Sigma_{r0,r1} arg0[r0,r1,s0,s1]*arg1[r0,r1,s2]`

    or

    `out[s0,s1]=Sigma_{r0,r1} arg0[r0,r1,s0,s1]*arg1[r0,r1]`

    In the first case the first dimension of ``arg0`` and the first dimension of
    ``arg1`` must match and in the second case the two first dimensions of
    ``arg0`` must match the two first dimensions of ``arg1``.

    The function call ``transposed_tensor_mult(arg0,arg1)`` is equivalent to
    ``tensor_mult(transpose(arg0),arg1)``.

    :param arg0: first argument of rank 2 or 4
    :type arg0: ``numpy.ndarray``, `escript.Data`, `Symbol`
    :param arg1: second argument of shape greater of 1 or 2 depending on the
                 rank of ``arg0``
    :type arg1: ``numpy.ndarray``, `escript.Data`, `Symbol`
    :return: the tensor product of transpose of arg0 and arg1 at each data point
    :rtype: ``numpy.ndarray``, `escript.Data`, `Symbol` depending on the input
    """
    from . import util
    sh0 = util.getShape(arg0)
    sh1 = util.getShape(arg1)
    if len(sh0) == 2 and (len(sh1) == 2 or len(sh1) == 1):
        return generalTransposedTensorProduct(arg0, arg1, axis_offset=1)
    elif len(sh0) == 4 and (len(sh1) == 2 or len(sh1) == 3 or len(sh1) == 4):
        return generalTransposedTensorProduct(arg0, arg1, axis_offset=2)
    else:
        raise ValueError("first argument must have rank 2 or 4")


def generalTransposedTensorProduct(arg0, arg1, axis_offset=0):
    """
    Generalized tensor product of transposed of ``arg0`` and ``arg1``.

    `out[s,t]=Sigma_r arg0[r,s]*arg1[r,t]`

    where
        - s runs through ``arg0.Shape[axis_offset:]``
        - r runs through ``arg0.Shape[:axis_offset]``
        - t runs through ``arg1.Shape[axis_offset:]``

    The function call ``generalTransposedTensorProduct(arg0,arg1,axis_offset)``
    is equivalent to
    ``generalTensorProduct(transpose(arg0,arg0.ndim-axis_offset),arg1,axis_offset)``.

    :param arg0: first argument
    :type arg0: ``numpy.ndarray``, `escript.Data`, `Symbol`, ``float``, ``int``
    :param arg1: second argument
    :type arg1: ``numpy.ndarray``, `escript.Data`, `Symbol`, ``float``, ``int``
    :return: the general tensor product of ``transpose(arg0)`` and ``arg1`` at
             each data point
    :rtype: ``numpy.ndarray``, `escript.Data`, `Symbol` depending on the input
    """
    from . import util
    if (isinstance(arg0, float) and isinstance(arg1, float)) or (isinstance(arg0, complex) and isinstance(arg1, complex)):
        return arg1 * arg0
    arg0, arg1 = util.matchType(arg0, arg1)
    if isinstance(arg0, sym.Symbol):
        sh0 = arg0.getShape()
        sh1 = util.getShape(arg1)
        if not sh0[:axis_offset] == sh1[:axis_offset]:
            raise ValueError("dimensions of last %s components in left argument don't match the first %s components in the right argument." % (axis_offset, axis_offset))
        if isinstance(arg1, float):
            return arg0 * arg1
        elif isinstance(arg1, numpy.ndarray) or isinstance(arg1, sym.Symbol):
            return arg0.transposedTensorProduct(arg1, axis_offset)
        elif isinstance(arg1, escore.Data):
            raise TypeError("tensor product of Symbol and Data not supported yet")
    elif isinstance(arg0, numpy.ndarray):
        if not arg0.shape[:axis_offset] == arg1.shape[:axis_offset]:
            raise ValueError("dimensions of last %s components in left argument don't match the first %s components in the right argument." % (axis_offset, axis_offset))
        arg0_c = arg0.copy()
        arg1_c = arg1.copy()
        sh0, sh1 = arg0.shape, arg1.shape
        d0, d1, d01 = 1, 1, 1
        for i in sh0[axis_offset:]:
            d0 *= i
        for i in sh1[axis_offset:]:
            d1 *= i
        for i in sh0[:axis_offset]:
            d01 *= i
        arg0_c.resize((d01, d0))
        arg1_c.resize((d01, d1))
        target_type = arg0.dtype if arg0.dtype.kind == 'c' else arg1.dtype
        out = numpy.zeros((d0, d1), target_type)
        for i0 in range(d0):
            for i1 in range(d1):
                out[i0, i1] = numpy.sum(arg0_c[:, i0] * arg1_c[:, i1])
        out.resize(sh0[axis_offset:] + sh1[axis_offset:])
        return out
    elif isinstance(arg0, escore.Data):
        if isinstance(arg1, sym.Symbol):
            raise TypeError("tensor product of Data and Symbol not supported yet")
        return escript_generalTransposedTensorProduct(arg0, arg1, axis_offset)


def escript_generalTransposedTensorProduct(arg0, arg1, axis_offset):
    """
    Generalized tensor product of transposed arg0 and arg1 for escript Data objects.

    :param arg0: first Data object (will be transposed)
    :type arg0: `escript.Data`
    :param arg1: second Data object (not necessarily on the same function space)
    :type arg1: `escript.Data`
    :param axis_offset: axis offset for the tensor product
    :type axis_offset: ``int``
    :return: tensor product of transpose(arg0) and arg1
    :rtype: `escript.Data`
    """
    return C_GeneralTensorProduct(arg0, arg1, axis_offset, 1)


def matrix_transposed_mult(arg0, arg1):
    """
    matrix-transposed(matrix) product of the two arguments.

    `out[s0,s1]=Sigma_{r0} arg0[s0,r0]*arg1[s1,r0]`

    The function call ``matrix_transposed_mult(arg0,arg1)`` is equivalent to
    ``matrix_mult(arg0,transpose(arg1))``.

    The last dimensions of ``arg0`` and ``arg1`` must match.

    :param arg0: first argument of rank 2
    :type arg0: ``numpy.ndarray``, `escript.Data`, `Symbol`
    :param arg1: second argument of rank 1 or 2
    :type arg1: ``numpy.ndarray``, `escript.Data`, `Symbol`
    :return: the product of ``arg0`` and the transposed of ``arg1`` at each data
             point
    :rtype: ``numpy.ndarray``, `escript.Data`, `Symbol` depending on the input
    :raise ValueError: if the shapes of the arguments are not appropriate
    """
    from . import util
    sh0 = util.getShape(arg0)
    sh1 = util.getShape(arg1)
    if not len(sh0) == 2:
        raise ValueError("first argument must have rank 2")
    if not len(sh1) == 2 and not len(sh1) == 1:
        raise ValueError("second argument must have rank 1 or 2")
    return generalTensorTransposedProduct(arg0, arg1, axis_offset=1)


def tensor_transposed_mult(arg0, arg1):
    """
    The tensor product of the first and the transpose of the second argument.

    For ``arg0`` of rank 2 this is

    `out[s0,s1]=Sigma_{r0} arg0[s0,r0]*arg1[s1,r0]`

    and for ``arg0`` of rank 4 this is

    `out[s0,s1,s2,s3]=Sigma_{r0,r1} arg0[s0,s1,r0,r1]*arg1[s2,s3,r0,r1]`

    or

    `out[s0,s1,s2]=Sigma_{r0,r1} arg0[s0,s1,r0,r1]*arg1[s2,r0,r1]`

    In the first case the second dimension of ``arg0`` and ``arg1`` must
    match and in the second case the two last dimensions of ``arg0`` must match
    the two last dimensions of ``arg1``.

    The function call ``tensor_transpose_mult(arg0,arg1)`` is equivalent to
    ``tensor_mult(arg0,transpose(arg1))``.

    :param arg0: first argument of rank 2 or 4
    :type arg0: ``numpy.ndarray``, `escript.Data`, `Symbol`
    :param arg1: second argument of shape greater of 1 or 2 depending on rank
                 of ``arg0``
    :type arg1: ``numpy.ndarray``, `escript.Data`, `Symbol`
    :return: the tensor product of ``arg0`` and the transposed of ``arg1`` at
             each data point
    :rtype: ``numpy.ndarray``, `escript.Data`, `Symbol` depending on the input
    """
    from . import util
    sh0 = util.getShape(arg0)
    sh1 = util.getShape(arg1)
    if len(sh0) == 2 and (len(sh1) == 2 or len(sh1) == 1):
        return generalTensorTransposedProduct(arg0, arg1, axis_offset=1)
    elif len(sh0) == 4 and (len(sh1) == 2 or len(sh1) == 3 or len(sh1) == 4):
        if len(sh1) == 2:
            return generalTensorTransposedProduct(arg0, transpose(arg1), axis_offset=2)
        else:
            return generalTensorTransposedProduct(arg0, arg1, axis_offset=2)
    else:
        raise ValueError("first argument must have rank 2 or 4")


def generalTensorTransposedProduct(arg0, arg1, axis_offset=0):
    """
    Generalized tensor product of ``arg0`` and transpose of ``arg1``.

    `out[s,t]=Sigma_r arg0[s,r]*arg1[t,r]`

    where
        - s runs through ``arg0.Shape[:arg0.ndim-axis_offset]``
        - r runs through ``arg0.Shape[arg1.ndim-axis_offset:]``
        - t runs through ``arg1.Shape[arg1.ndim-axis_offset:]``

    The function call ``generalTensorTransposedProduct(arg0,arg1,axis_offset)``
    is equivalent to
    ``generalTensorProduct(arg0,transpose(arg1,arg1.ndim-axis_offset),axis_offset)``.

    :param arg0: first argument
    :type arg0: ``numpy.ndarray``, `escript.Data`, `Symbol`, ``float``, ``int``
    :param arg1: second argument
    :type arg1: ``numpy.ndarray``, `escript.Data`, `Symbol`, ``float``, ``int``
    :return: the general tensor product of ``arg0`` and ``transpose(arg1)`` at
             each data point
    :rtype: ``numpy.ndarray``, `escript.Data`, `Symbol` depending on the input
    """
    from . import util
    if ((isinstance(arg0, float) or isinstance(arg0, complex)) and
            (isinstance(arg1, float) or isinstance(arg1, complex))):
        return arg1 * arg0
    arg0, arg1 = util.matchType(arg0, arg1)
    if isinstance(arg0, sym.Symbol):
        sh0 = arg0.getShape()
        sh1 = util.getShape(arg1)
        r1 = util.getRank(arg1)
        if not sh0[arg0.getRank() - axis_offset:] == sh1[r1 - axis_offset:]:
            raise ValueError("dimensions of last %s components in left argument don't match the first %s components in the right argument." % (axis_offset, axis_offset))
        if isinstance(arg1, float) or isinstance(arg1, complex):
            return arg0 * arg1
        elif isinstance(arg1, numpy.ndarray) or isinstance(arg1, sym.Symbol):
            return arg0.tensorTransposedProduct(arg1, axis_offset)
        elif isinstance(arg1, escore.Data):
            raise TypeError("tensor product of Symbol and Data not supported yet")
    elif isinstance(arg0, numpy.ndarray):
        if not (arg0.shape[arg0.ndim - axis_offset:] == arg1.shape[arg1.ndim - axis_offset:] or
                arg0.shape[arg0.ndim - axis_offset:], tuple(reversed(arg1.shape[arg1.ndim - axis_offset:]))):
            raise ValueError("dimensions of last %s components in left argument don't match the first %s components in the right argument." % (axis_offset, axis_offset))
        arg0_c = arg0.copy()
        arg1_c = arg1.copy()
        sh0, sh1 = arg0.shape, arg1.shape
        d0, d1, d01 = 1, 1, 1
        for i in sh0[:arg0.ndim - axis_offset]:
            d0 *= i
        for i in sh1[:arg1.ndim - axis_offset]:
            d1 *= i
        for i in sh1[arg1.ndim - axis_offset:]:
            d01 *= i
        arg0_c.resize((d0, d01))
        arg1_c.resize((d1, d01))
        if arg0_c.dtype != numpy.double:
            out = numpy.zeros((d0, d1), arg0_c.dtype)
        else:
            out = numpy.zeros((d0, d1), numpy.double)
        for i0 in range(d0):
            for i1 in range(d1):
                out[i0, i1] = numpy.sum(arg0_c[i0, :] * arg1_c[i1, :])
        out.resize(sh0[:arg0.ndim - axis_offset] + sh1[:arg1.ndim - axis_offset])
        return out
    elif isinstance(arg0, escore.Data):
        if isinstance(arg1, sym.Symbol):
            raise TypeError("tensor product of Data and Symbol not supported yet")
        return escript_generalTensorTransposedProduct(arg0, arg1, axis_offset)


def escript_generalTensorTransposedProduct(arg0, arg1, axis_offset):
    """
    Generalized tensor product of arg0 and transposed arg1 for escript Data objects.

    :param arg0: first Data object
    :type arg0: `escript.Data`
    :param arg1: second Data object (will be transposed, not necessarily on the same function space)
    :type arg1: `escript.Data`
    :param axis_offset: axis offset for the tensor product
    :type axis_offset: ``int``
    :return: tensor product of arg0 and transpose(arg1)
    :rtype: `escript.Data`
    """
    return C_GeneralTensorProduct(arg0, arg1, axis_offset, 2)
